import re
import typing as t
import uuid

import click

from globus_sdk.paging import Paginator

from gcs_cli.helpers import default_identity_id_resolver, requires_login
from gcs_cli.helpers.field_funcs import FieldSpec
from gcs_cli.parsing import globus_command, include_private_policies_opt
from gcs_cli.prettyprint import FORMAT_TEXT_TABLE, formatted_print
from gcs_cli.services.gcs import get_client


def get_last_access(obj: t.Dict[str, t.Any]) -> str:
    last_access = obj.get("last_access")
    if last_access is None:
        data_type = obj.get("DATA_TYPE", "")
        version_match = re.match(
            r"collection#(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<patch>[0-9]+)",
            data_type,
        )
        if version_match is None or (
            int(version_match.group("major")) == 1
            and int(version_match.group("minor")) < 9
        ):
            return ""
        return "never"
    return str(last_access)


STANDARD_FIELDS: FieldSpec = [
    ("ID", "id"),
    ("Display Name", "display_name"),
    ("Owner", default_identity_id_resolver.field),
    ("Collection Type", "collection_type"),
    ("Storage Gateway ID", "storage_gateway_id"),
    ("Created", lambda obj: obj.get("created_at", "")),
    ("Last Access", get_last_access),
]


class FilterType(click.ParamType):
    """
    Let people use "mapped-collections" or "mapped_collections" as a nicety.
    Also parse the created_at <=> date and last_access <=> date options
    """

    name = "filter"
    basic_filters = [
        "mapped_collections",
        "guest_collections",
        "managed_by_me",
        "created_by_me",
    ]
    filters = basic_filters + [
        r"created_at\s*([<>]=?|=)\s*\d{4}-\d{2}-\d{2}",
        r"last_access\s*([<>]=?|=)\s*\d{4}-\d{2}-\d{2}",
    ]
    filter_re = re.compile("^(" + "|".join(filters) + ")$")

    @staticmethod
    def hyphen_to_underscore(value: str) -> str:
        def dehyphen(s: re.Match[str]) -> str:
            return s.group(1).replace("-", "_")

        return re.sub("^([a-z_-]*)", dehyphen, value).replace(" ", "")

    def convert(
        self,
        value: str,
        param: t.Optional[click.Parameter],
        ctx: t.Optional[click.Context],
    ) -> t.Optional[str]:
        if value is None:
            return None

        value = self.hyphen_to_underscore(value)
        if not FilterType.filter_re.match(value):
            raise click.BadParameter(
                f"Unable to parse filter {value}", ctx=ctx, param=param
            )

        return t.cast(str, super().convert(value, param, ctx))


@globus_command("list", short_help="List Collections")
@include_private_policies_opt
@requires_login
@click.option(
    "--mapped-collection-id",
    default=None,
    type=click.UUID,
    help=(
        "Filter results to Guest Collections on a specific Mapped Collection. "
        "This is the ID of the Mapped Collection"
    ),
)
@click.option(
    "--filter",
    "filters",
    multiple=True,
    # choices are shown with hyphens, but the command will receive them with underscores
    type=FilterType(),
    help="""\
Filter results to one of the specified categories of collections. Can be applied
multiple times. Note that mutually exclusive filters are allowed and will find no
results.

The filters are as follows

\b
mapped-collections:
  Only show collections with collection_type="mapped"

\b
guest-collections:
  Only show collections with collection_type="guest"

\b
managed-by-me:
  Only show collections where one of your identities has a role

\b
created-by-me:
  Only show collections where one of your identities was the creator

\b
created-at [< | <= | = | => | >] YYYY-MM-DD:
  Only show collections which were created before, after, or on a certain date.

\b
last-access [< | <= | = | => | >] YYYY-MM-DD:
  Only show collections which have been last used before, after, or on a
  certain date.
""",
)
def collection_list(
    include_private_policies: bool,
    mapped_collection_id: uuid.UUID,
    filters: t.Sequence[str],
) -> None:
    """
    List Collections on the current Endpoint
    """
    client = get_client()
    params = {}
    if mapped_collection_id:
        params["mapped_collection_id"] = str(mapped_collection_id)
    if filters:
        params["filter"] = ",".join(filters)
    if include_private_policies:
        params["include"] = "private_policies"

    res: Paginator = client.paginated.get_collection_list(**params)

    formatted_print(res, text_format=FORMAT_TEXT_TABLE, fields=STANDARD_FIELDS)
