import uuid
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import click

from globus_sdk import GCSAPIError

from gcs_cli.helpers import record_print_result_message, requires_login
from gcs_cli.parsing import EXPLICIT_NULL, IntRangeOrNull, globus_command, node_id_arg
from gcs_cli.parsing.types import _ExplicitNullClass
from gcs_cli.sdk_client.data import NodeDocument
from gcs_cli.services.gcs import get_client


def validate_port_range(
    ctx: click.Context,
    param: click.Parameter,
    value: Optional[Sequence[Union[int, _ExplicitNullClass]]],
) -> Union[_ExplicitNullClass, Tuple[int, int], Sequence]:
    if value is None:
        return []

    if len(value) == 0:
        return []

    _min, _max = value
    if _min == EXPLICIT_NULL or _max == EXPLICIT_NULL:
        if (_min == EXPLICIT_NULL) != (_max == EXPLICIT_NULL):
            raise click.BadParameter('Ports must both be integers or ""')
        return EXPLICIT_NULL

    assert isinstance(_min, int)
    assert isinstance(_max, int)

    if _min > _max:
        raise click.BadParameter("LOW_PORT must be <= HIGH_PORT")
    return value


@globus_command("update", short_help="Update a Node's config, including enable/disable")
@node_id_arg
@click.option(
    "--enable/--disable",
    default=None,
    help="Set the node state to 'active' or 'inactive'",
)
@click.option(
    "--ip-address",
    "-i",
    "ip_addresses",
    multiple=True,
    help=(
        "IP address of the GCS Node. Use this option multiple times to set multiple IPs"
    ),
)
@click.option(
    "--data-interface",
    "data_interface",
    help=("IP interface of the GCS Node used for Globus data transfers"),
)
@click.option(
    "--incoming-port-range",
    help=(
        "Allowed port range for incoming TCP data connections. "
        'Set to "" "" to remove this setting'
    ),
    metavar="LOW_PORT HIGH_PORT",
    type=IntRangeOrNull(1, 2**16 - 1),
    nargs=2,
    callback=validate_port_range,
)
@click.option(
    "--outgoing-port-range",
    help=(
        "Port range used as the source for outgoing TCP data connections. "
        'Set to "" "" to remove this setting'
    ),
    metavar="LOW_PORT HIGH_PORT",
    type=IntRangeOrNull(1, 2**16 - 1),
    nargs=2,
    callback=validate_port_range,
)
@requires_login
def node_update(
    node_id: uuid.UUID,
    enable: bool,
    ip_addresses: Sequence[str],
    data_interface: Optional[str],
    incoming_port_range: Optional[Union[_ExplicitNullClass, Tuple[int, int]]],
    outgoing_port_range: Optional[Union[_ExplicitNullClass, Tuple[int, int]]],
) -> None:
    """
    Update a Node on the current Endpoint.
    """
    data: Dict[str, Any] = {"id": str(node_id)}
    if enable is not None:
        data["status"] = "active" if enable else "inactive"
    if ip_addresses:
        data["ip_addresses"] = list(ip_addresses)

    if data_interface:
        data["data_interface"] = data_interface
    elif len(ip_addresses) == 1:
        data["data_interface"] = ip_addresses[0]
    elif len(ip_addresses) != 0:
        raise click.exceptions.UsageError(
            "You must specificy --data-interface when "
            "using multiple --ip-address values"
        )

    if incoming_port_range == EXPLICIT_NULL:
        data["incoming_port_range"] = None
    elif incoming_port_range:
        assert not isinstance(incoming_port_range, _ExplicitNullClass)
        data["incoming_port_range"] = list(incoming_port_range)

    if outgoing_port_range == EXPLICIT_NULL:
        data["outgoing_port_range"] = None
    elif outgoing_port_range:
        assert not isinstance(outgoing_port_range, _ExplicitNullClass)
        data["outgoing_port_range"] = list(outgoing_port_range)
    doc = NodeDocument(**data)

    client = get_client()
    try:
        res = client.update_node(node_id, doc)
    except GCSAPIError as e:
        if enable and e.http_status == 503:
            click.echo(
                (
                    "Node enable failed with an error which may indicate that you are "
                    "trying to enable the current host after it has been disabled. "
                    "This is not possible."
                    "\n\nYou should re-run"
                    "\n  globus-connect-server node setup\n"
                    "instead."
                ),
                err=True,
            )
        raise
    record_print_result_message(res)
