import errno
import json
import socket
import ssl
import textwrap
import traceback
import urllib.parse
from types import TracebackType
from typing import Any, Iterable, List, Mapping, Optional, Tuple, Type, Union

import click
import requests
import urllib3.exceptions

import globus_sdk
from globus_sdk import AuthAPIError, GCSAPIError

from gcs_cli.parsing.stateful_options import get_show_stack_trace, outformat_is_text
from gcs_cli.services.auth import get_client


def handle_excinfo(
    excinfo: Tuple[Type[BaseException], BaseException, TracebackType]
) -> None:
    exception_type, exception, _ = excinfo

    if get_show_stack_trace():
        click.echo("".join(traceback.format_exception(*excinfo)), err=True)
    if isinstance(
        exception, (click.ClickException, click.exceptions.Abort, click.exceptions.Exit)
    ):
        raise exception
    elif isinstance(exception, GCSAPIError):
        _gcsapi_error_hook(exception)
    elif isinstance(exception, AuthAPIError):
        _authapi_error_hook(exception)
    elif isinstance(exception, globus_sdk.GlobusAPIError):
        _default_api_error_hook(exception)
    elif isinstance(exception, globus_sdk.GlobusConnectionError):
        _default_connection_error_hook(exception)
    else:
        click.echo("{}: {}".format(exception_type.__name__, exception))
    click.get_current_context().exit(1)


def _default_connection_error_hook(exception: globus_sdk.GlobusConnectionError) -> None:
    click.secho("Error contacting the GCS Manager API", err=True, fg="red", bold=True)

    requests_exc = exception.underlying_exception
    netloc = None
    if requests_exc and isinstance(requests_exc, requests.exceptions.ConnectionError):
        try:
            request_url = requests_exc.request.url
            netloc = urllib.parse.urlparse(request_url).netloc
        except Exception:
            pass

        for arg in requests_exc.args:
            if isinstance(arg, urllib3.exceptions.RequestError):
                context = getattr(arg, "__context__", None)
                while context:
                    if isinstance(context, (socket.gaierror, ConnectionError)):
                        _handle_network_error(netloc, context)
                        context = None
                    elif isinstance(context, ssl.SSLError):
                        _handle_ssl_error(netloc, context)
                        context = None
                    else:
                        context = getattr(context, "__context__", None)


def _handle_network_error(
    netloc: Optional[str], exception: Union[socket.gaierror, ConnectionError]
) -> None:
    if exception.errno == socket.EAI_NONAME:
        click.secho(
            "\n".join(
                ["Error resolving {}".format(netloc)]
                + textwrap.wrap(
                    "This may be because the endpoint is deleted, it is not deployed "
                    "on any nodes, or your DNS resolver is misconfigured."
                )
            ),
            err=True,
        )
    elif exception.errno == errno.ECONNREFUSED:
        click.secho(
            "\n".join(
                ["Error connecting to {}".format(netloc)]
                + textwrap.wrap(
                    "This may be because the GCS services are not running on the "
                    "endpoint nodes, or an endpoint node's IP address has changed."
                )
            ),
            err=True,
        )
    else:
        click.secho(
            "\n".join(
                textwrap.wrap(
                    "Error connecting to {}: {}".format(netloc, str(exception))
                )
            ),
            err=True,
        )


def _handle_ssl_error(netloc: Optional[str], exception: ssl.SSLError) -> None:
    if exception.reason == "CERTIFICATE_VERIFY_FAILED":
        click.secho(
            "\n".join(
                ["SSL error connecting to {}: {}".format(netloc, exception.reason)]
                + textwrap.wrap(
                    "This may be because the trust roots for your system don't support "
                    "the Let's Encrypt CA or another virtual host is configured to use "
                    "the same server name as the GCS Manager."
                )
            ),
            err=True,
        )
    else:
        click.secho(
            "SSL error connecting to {}: {}".format(netloc, exception.reason),
            err=True,
        )


def _default_api_error_hook(exception: globus_sdk.GlobusAPIError) -> None:
    click.secho("An API Error Occurred", err=True, fg="red", bold=True)
    click.secho("HTTP status: {}".format(exception.http_status), err=True, fg="yellow")
    click.echo("code:        {}".format(exception.code), err=True)
    if exception.message:
        click.echo("message:     {}".format(exception.message), err=True)


def _auth_timeout_error_hook(exception: Union[GCSAPIError, AuthAPIError]) -> None:
    click.secho("An authentication timeout occurred", err=True, fg="red", bold=True)

    ids: List[str] = []
    require_mfa: bool = False

    if isinstance(exception, GCSAPIError):
        if isinstance(exception.detail, Mapping):
            click.echo(
                "This means you would have been authorized based on your identities, "
                "but you haven't logged in recently enough.",
                err=True,
            )
            ids = exception.detail["identities"]
            require_mfa = exception.detail.get("require_mfa", False)
        else:
            return
    elif isinstance(exception, AuthAPIError):
        details = json.loads(exception.raw_json["errors"][0]["detail"])  # type: ignore
        click.echo(details["authorization_parameters"]["session_message"], err=True)
        ids = details["authorization_parameters"]["session_required_identities"]

    if len(ids) > 1:
        cmdmsg = "To resolve, you should run one of these commands"
    else:
        cmdmsg = "To resolve, you should run this command"

    click.secho(cmdmsg + " to update your authentication:", err=True, fg="yellow")
    mfa_flag = ""
    if require_mfa:
        mfa_flag = "--mfa "
    for n in _try_resolve_ids_to_usernames(ids):
        click.echo(
            "  globus-connect-server session update {}{}".format(mfa_flag, n),
            err=True,
        )


def _required_scopes_error_hook(exception: GCSAPIError) -> None:
    click.secho(
        "This operation requires additional authentication.", err=True, fg="yellow"
    )
    if isinstance(exception.detail, Mapping):
        scopes = exception.detail.get("required_scopes")
        if scopes:
            click.secho("Update your consents with:\n", err=True)
            click.echo(
                "  globus-connect-server session consent {}".format(" ".join(scopes)),
                err=True,
            )


def _json_api_error_hook(exception: Union[GCSAPIError, AuthAPIError]) -> None:
    body: Any
    if exception.raw_json:
        body = exception.raw_json
    else:
        body = exception.raw_text
    response = exception._underlying_response
    headers = dict(response.headers)
    gcs_headers = {k: v for k, v in headers.items() if k.startswith("x-gcs-")}
    other_headers = {k: v for k, v in headers.items() if k not in gcs_headers}
    click.secho(
        json.dumps(
            {
                "is_error": True,
                "http_status": exception.http_status,
                "code": exception.code,
                "message": exception.message,
                "body": body,
                "gcs_headers": gcs_headers,
                "headers": other_headers,
            },
            indent=2,
            separators=(",", ": "),
            sort_keys=True,
        ),
        fg="yellow",
    )


def _gcsapi_error_hook(exception: GCSAPIError) -> None:
    if outformat_is_text():
        if exception.detail_data_type == "authentication_timeout#1.1.0":
            _auth_timeout_error_hook(exception)
        elif exception.detail_data_type == "missing_required_scopes#1.0.0":
            _required_scopes_error_hook(exception)
        else:
            _default_api_error_hook(exception)
    else:
        _json_api_error_hook(exception)


def _authapi_error_hook(exception: AuthAPIError) -> None:
    if outformat_is_text():
        if exception.http_status == 403:
            if "authorization_parameters" in exception.message:
                try:
                    _auth_timeout_error_hook(exception)
                    return
                except Exception:
                    # If Auth response format changes, let's ask the user to
                    # notify us and then fall through to the default handler.
                    click.secho(
                        "An exception has occurred while handling the response below "
                        "Please report this to support@globus.org.\n\n"
                        f"===========================================================\n"
                        f"Error while processing {exception}\n"
                        f"===========================================================",
                        err=True,
                        fg="red",
                        bold=True,
                    )
        _default_api_error_hook(exception)
    else:
        _json_api_error_hook(exception)


def _try_resolve_ids_to_usernames(ids: Iterable[str]) -> Iterable[str]:
    """resolve IDs to usernames, falling back to the IDs which were given"""
    try:
        idmap = globus_sdk.IdentityMap(get_client())
        # add all to do batched lookups
        for i in ids:
            idmap.add(i)
        # do the lookups and produce the usernames
        return [idmap[i]["username"] for i in ids]
    # very conservative error handling -- if anything goes wrong, just return the
    # original IDs
    except Exception:
        return ids
