from typing import cast, Dict, List, Optional, Set, Tuple, TYPE_CHECKING

import click

from globus_sdk import GlobusError, IdentityMap

from ..models import GridmapConfiguration, MigrationPlan, S3Credential
from ..util import get_api
from .decorators import requires_logged_in, requires_migration_plan
from ._fixup_s3_credential import _fixup_s3_credential

if TYPE_CHECKING:
    from ..typing.gcs_documents import (  # noqa: F401
        ExpressionIdentityMappingDict,
        IdentityMappingDict,
    )


@click.command("gridmap", help="Add Authorization for gridmap entries")
@click.option(
    "--full-mapping",
    "-f",
    multiple=True,
    nargs=3,
    metavar="SUBJECT GLOBUS_USERNAME LOCAL_USERNAME",
)
@click.option(
    "--subject-mapping", "-s", multiple=True, nargs=2, metavar="SUBJECT GLOBUS_USERNAME"
)
@click.option(
    "--dont-update-guest-collections",
    "-D",
    is_flag=True,
    help="If specified, don't update guest collection configurations to use "
    "this mapping.",
)
@requires_logged_in
@requires_migration_plan
def identity_mapping_gridmap(
    full_mapping: Tuple[Tuple[str, str, str]],
    subject_mapping: Tuple[Tuple[str, str]],
    dont_update_guest_collections: bool,
    migration_plan: MigrationPlan,
) -> None:
    failed = 0
    full_mapping_list = list(full_mapping)

    if not migration_plan.gsi_authz_configuration:
        raise click.ClickException("Migration plan has no GSI configuration")
    if not migration_plan.gsi_authz_configuration.callouts:
        raise click.ClickException("Migration plan has no callouts configured")
    if not migration_plan.storage_gateway:
        raise click.ClickException("Migration plan has no storage gateway")

    gridmap_configs: Dict[int, GridmapConfiguration] = {}

    for idx, config in enumerate(migration_plan.gsi_authz_configuration.callouts):
        if config.is_gridmap:
            gridmap_configs[idx] = cast(GridmapConfiguration, config)

    if not gridmap_configs:
        raise click.ClickException("No gridmap entries configured for this endpoint")

    allowed_domains: Set[str] = set()
    identity_mappings: List["ExpressionIdentityMappingDict"] = []
    # reverse map of local usernames to globus usernames
    reverse_map: Dict[str, str] = {}

    # For subject mappings, look up entries in gridmap and add the results
    # to full_mappings_list
    for sm in subject_mapping:
        subject, globus_username = sm
        local_usernames: Set[str] = set()

        for gridmap_config in gridmap_configs.values():
            mapped_subject = False
            if gridmap_config.mappings:
                mapped_local = gridmap_config.mappings.get(subject)
                if not mapped_local:
                    continue
                local_usernames.update(mapped_local)

        if len(local_usernames) == 0:
            if not click.prompt(
                f"{subject} does not have a mapping defined. Use "
                "-f to choose which users to map. Continue with other "
                "mappings?",
            ):
                raise click.ClickException(
                    f"Incomplete mapping information for {subject}"
                )
            failed += 1
            continue

        if len(local_usernames) > 1:
            if not click.prompt(
                f"{subject} maps to multiple users: {mapped_local}. Use "
                "-f to choose which users to map. Continue with other "
                "mappings?",
            ):
                raise click.ClickException("Ambiguous mapping information.")
            failed += 1
            continue

        full_mapping_entry = (subject, globus_username, local_usernames.pop())
        full_mapping_list.append(full_mapping_entry)

    # Create mappings for users
    for fm in full_mapping_list:
        subject, globus_username, local_username = fm
        mapped_subject = False
        identity_mapping: "ExpressionIdentityMappingDict" = {
            "DATA_TYPE": "expression_identity_mapping#1.0.0",
            "mappings": [
                {
                    "source": "{username}",
                    "match": f"{globus_username}",
                    "output": f"{local_username}",
                    "literal": True,
                }
            ],
        }

        for gridmap_config in gridmap_configs.values():
            if gridmap_config.mappings:
                mapped_local = gridmap_config.mappings.get(subject)
                if not mapped_local:
                    continue
                if local_username in mapped_local:
                    mapped_subject = True
                gridmap_config.add_domain(globus_username.rsplit("@", 1)[1])
        if not mapped_subject:
            click.echo(
                f"There is no mapping from {subject} to {local_username} in the "
                "gridmap configuration. Use gcs-migration map-identities to "
                "add other user mappings."
            )
            failed += 1
            continue
        identity_mappings.append(identity_mapping)
        reverse_map[local_username] = globus_username

    # Look up Globus users to figure out which globus usernames are valid
    identity_map = IdentityMap(get_api("auth.globus.org"))
    for idm in identity_mappings:
        identity_map.add(idm["mappings"][0]["match"])

    valid_identity_mappings: List["ExpressionIdentityMappingDict"] = []

    for idm in identity_mappings:
        try:
            identity = identity_map[idm["mappings"][0]["match"]]
            valid_identity_mappings.append(idm)
            if identity["identity_type"] != "email":
                domain = idm["mappings"][0]["match"].rsplit("@", 1)[1]
                allowed_domains.add(domain)
        except (GlobusError, KeyError):
            click.echo(
                f"Unable to resolve {idm['mappings'][0]['match']} to a valid "
                f"Globus identity, ignoring mapping to {idm['mappings'][0]['output']}."
            )
            failed += 1

    # Add domains and mappings to storage gateway
    migration_plan.storage_gateway.add_allowed_domains(
        allowed_domains, cast("List[IdentityMappingDict]", valid_identity_mappings)
    )

    if migration_plan.guest_collections or migration_plan.credentials:
        # Update the credentials which match the gridmap entries
        if dont_update_guest_collections:
            click.echo(
                "Not updating guest collection ownership. Use gcs-migration "
                "guest-collection set-owner to update guest collections with "
                "appropriate ownership."
            )
            return

        for cred in migration_plan.credentials:
            if cred and cred.gsi_authz_callout_index in gridmap_configs.keys():
                if cred.identity_username is None:
                    identity_username: Optional[str] = None
                    if cred.username:
                        identity_username = reverse_map.get(cred.username)
                    if not identity_username:
                        continue
                    try:
                        identity = identity_map[identity_username]
                        cred.identity_username = identity_username
                        cred.identity_id = identity["id"]

                        # For S3 connector the identity mapping is a no-op,
                        # but migration-plan create doesn't know about domains.
                        # If this is a new domain assignment for this credential,
                        # update the local username to match the globus username.
                        if isinstance(cred, S3Credential):
                            _fixup_s3_credential(migration_plan, cred)
                    except KeyError:
                        click.echo(
                            f"Unable to resolve {identity_username} to a valid "
                            "Globus identity, not updating configuration for "
                            f"{cred}"
                        )
                        failed += 1
    migration_plan.dump_storage_gateway()
    migration_plan.dump_credentials()
    migration_plan.dump_guest_collections()
    migration_plan.dump_gsi_authz_configuration()
    click.get_current_context().exit(failed)
