import os
from typing import Any, cast, Dict, List, Optional, Union
from typing_extensions import TypedDict

from ..util import yaml, yaml_object
from ..parsers import Gridmap


class GridmapConfigurationDict(TypedDict):
    gridmap: Optional[Union[str]]
    mappings: Optional[Dict[str, List[str]]]
    domains: List[str]


@yaml_object(yaml)
class GridmapConfiguration:
    def __init__(
        self, gridmap: "Optional[Union[os.PathLike[str], str]]" = None
    ) -> None:
        self.mappings: Optional[Dict[str, List[str]]] = None
        self.gridmap: "Optional[Union[str, os.PathLike[str]]]" = gridmap
        self.domains: List[str] = []

    @property
    def is_myproxy(self) -> bool:
        return False

    @property
    def is_cilogon(self) -> bool:
        return False

    @property
    def is_gridmap(self) -> bool:
        return True

    @property
    def mappings(self) -> Optional[Dict[str, List[str]]]:
        return self._mappings

    @mappings.setter
    def mappings(self, value: Optional[Dict[str, List[str]]]) -> None:
        self._mappings = value

    @property
    def gridmap(self) -> "Optional[Union[str, os.PathLike[str]]]":
        return self._gridmap

    @gridmap.setter
    def gridmap(self, value: "Optional[Union[str, os.PathLike[str]]]") -> None:
        self._gridmap = value
        if self._mappings is None and value:
            parser = Gridmap(value)
            self.mappings = parser.mappings

    def add_domain(self, domain: str) -> None:
        if domain not in self.domains:
            self.domains.append(domain)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GridmapConfiguration):
            return NotImplemented
        if self is other:
            return True

        self_state = cast(Dict[str, Any], self.__getstate__())
        self_state.pop("domains", None)

        other_state = cast(Dict[str, Any], other.__getstate__())
        other_state.pop("domains", None)

        return self_state == other_state

    def __getstate__(self) -> GridmapConfigurationDict:
        if isinstance(self.gridmap, os.PathLike):
            gridmap: Optional[str] = self.gridmap.__fspath__()
        else:
            gridmap = self.gridmap

        return {
            "gridmap": gridmap,
            "mappings": self.mappings,
            "domains": self.domains,
        }

    def __setstate__(self, state: GridmapConfigurationDict) -> None:
        self.mappings = state["mappings"]
        self.gridmap = state["gridmap"]
        self.domains = state["domains"]
