import os
import pathlib
import shutil
from typing import List, Optional, Union
from typing import Any
from typing_extensions import TypedDict
import uuid

import packaging.version

from .credential import Credential
from .gsi_authz_configuration import GsiAuthzConfiguration
from .guest_collection import GuestCollection
from .mapped_collection import MappedCollection
from .node import Node
from .storage_gateway import StorageGateway

from ..parsers import GCSConfig, GridFTPOptionParser
from ..util import yaml, yaml_object
from ..version import (
    __version__,
    VERSION_INDEPENDENT_CREDENTIAL,
    VERSION_NO_VERSION_YAML,
)


class MigrationPlanNotFound(Exception):
    pass


class MigrationPlanUnsupportedVersion(Exception):
    def __init__(self, version: str) -> None:
        self.version = version


class MigrationPlanDict(TypedDict):
    environment: str
    version: Optional[str]
    gsi_authz_configuration: Optional[GsiAuthzConfiguration]
    storage_gateway: Optional[StorageGateway]
    mapped_collection: Optional[MappedCollection]
    guest_collections: Optional[List[GuestCollection]]
    credentials: List[Credential]


@yaml_object(yaml)
class MigrationPlan:
    def __init__(self, root: Optional[Union[str, os.PathLike]] = None) -> None:
        self.node: Optional[Node] = None
        self.gcs_config: Optional[GCSConfig] = None
        self.gridftp_config: Optional[GridFTPOptionParser] = None
        self.gsi_authz_configuration: Optional[GsiAuthzConfiguration] = None
        self.storage_gateway: Optional[StorageGateway] = None
        self.mapped_collection: Optional[MappedCollection] = None
        self.credentials: List[Credential] = []
        self.guest_collections: Optional[List[GuestCollection]] = None
        self.environment: Optional[str] = None
        self.version: Optional[str] = __version__
        self.root = pathlib.Path(root if root else ".")

    @property
    def environment(self) -> Optional[str]:
        if self._environment:
            return self._environment
        return None

    @environment.setter
    def environment(self, value: Optional[str]) -> None:
        self._environment = value

        if value and value.lower() != "default":
            os.environ["GLOBUS_SDK_ENVIRONMENT"] = value.lower()
        elif "GLOBUS_SDK_ENVIRONMENT" in os.environ:
            del os.environ["GLOBUS_SDK_ENVIRONMENT"]

    @property
    def gcs_config(self) -> Optional[GCSConfig]:
        return self._gcs_config

    @gcs_config.setter
    def gcs_config(self, value: Optional[GCSConfig]) -> None:
        self._gcs_config = value

        if value and not self.environment:
            self.environment = value.instance

    @property
    def gridftp_config(self) -> Optional[GridFTPOptionParser]:
        return self._gridftp_config

    @gridftp_config.setter
    def gridftp_config(self, value: Optional[GridFTPOptionParser]) -> None:
        self._gridftp_config = value

        if value and not self.gsi_authz_configuration:
            gsi_authz_conf = value.get_string_option("$GSI_AUTHZ_CONF")
            gridmap = value.get_string_option("$GRIDMAP")
            self.gsi_authz_configuration = GsiAuthzConfiguration(
                gsi_authz_conf=gsi_authz_conf,
                gridmap=gridmap,
                gridftp_config=value,
            )
        if value and not self.node:
            self.node = Node()
            self.node.gridftp_config = value

    def add_guest_collection(self, guest_collection: GuestCollection) -> None:
        if self.guest_collections is None:
            self.guest_collections = []
        self.guest_collections.append(guest_collection)
        if guest_collection.credential:
            self.add_credential(guest_collection.credential)

    def add_credential(self, credential: Credential) -> None:
        if credential and credential.id not in [cred.id for cred in self.credentials]:
            self.credentials.append(credential)

    def remove_guest_collection(self, guest_collection: GuestCollection) -> None:
        if self.guest_collections:
            self.guest_collections.remove(guest_collection)

    def get_guest_collection_by_id(
        self, value: Union[uuid.UUID, str]
    ) -> Optional[GuestCollection]:
        if isinstance(value, uuid.UUID):
            svalue = str(value)
        else:
            svalue = value
        if self.guest_collections:
            for g in self.guest_collections:
                if g.id == svalue:
                    return g
        return None

    @property
    def migration_plan_dir(self) -> pathlib.Path:
        return pathlib.Path(self.root) / "migration_plan"

    def dump(self, omit_secrets: bool = False) -> None:
        if os.path.exists(self.migration_plan_dir):
            shutil.rmtree(self.migration_plan_dir, ignore_errors=True)
        os.makedirs(self.migration_plan_dir, mode=0o755)

        version_file = self.migration_plan_dir / "version.yaml"
        with open(version_file, "w") as f:
            yaml.dump(__version__, f)

        environment_file = self.migration_plan_dir / "environment.yaml"
        with open(environment_file, "w") as f:
            yaml.dump(self.environment, f)

        self.dump_credentials(omit_secrets=omit_secrets)
        self.dump_guest_collections()
        self.dump_storage_gateway(omit_secrets=omit_secrets)
        self.dump_mapped_collection()

        if self.gsi_authz_configuration:
            self.dump_gsi_authz_configuration()

    def dump_gsi_authz_configuration(self) -> None:
        if self.gsi_authz_configuration:
            self.gsi_authz_configuration.dump(self.migration_plan_dir)

    def dump_credentials(self, omit_secrets: bool = False) -> None:
        cred_dir = self.migration_plan_dir / "credentials"
        if not os.path.exists(cred_dir):
            os.makedirs(cred_dir, mode=0o755)
        for cred_filename in os.listdir(cred_dir):
            os.remove(cred_dir / cred_filename)

        if self.credentials:
            for cred in self.credentials:
                cred.dump(cred_dir, omit_secrets=omit_secrets)

    def dump_guest_collection(self, id: str) -> None:
        guest_collection = self.get_guest_collection_by_id(id)
        if guest_collection:
            gcdir = self.migration_plan_dir / "guest_collections"
            guest_collection.dump(gcdir)

    def dump_guest_collections(self) -> None:
        if self.guest_collections:
            gcdir = self.migration_plan_dir / "guest_collections"
            if not os.path.exists(gcdir):
                os.makedirs(gcdir, mode=0o755)
            for gc in self.guest_collections:
                gc.dump(gcdir)

    def dump_mapped_collection(self) -> None:
        if self.mapped_collection:
            self.mapped_collection.dump(self.migration_plan_dir)

    def dump_storage_gateway(self, omit_secrets: bool = False) -> None:
        if self.storage_gateway:
            self.storage_gateway.dump(
                self.migration_plan_dir, omit_secrets=omit_secrets
            )

    @staticmethod
    def load(root: Optional[Union[str, os.PathLike]] = None) -> "MigrationPlan":
        root = pathlib.Path(root if root is not None else ".")

        if not os.path.exists(root / "migration_plan"):
            raise MigrationPlanNotFound(f"No migration plan found in {root}.")

        obj = MigrationPlan(root)

        # Older plans did not have versions
        load_version: str = VERSION_NO_VERSION_YAML

        version_file = obj.migration_plan_dir / "version.yaml"
        try:
            with open(version_file, "r") as f:
                load_version = yaml.load(f)
        except FileNotFoundError:
            pass
        except IOError:
            raise

        obj_version = str(load_version).replace("~", "")
        our_version = __version__.replace("~", "")
        if packaging.version.parse(obj_version) > packaging.version.parse(our_version):
            raise MigrationPlanUnsupportedVersion(obj_version)
        obj.version = __version__

        environment_file = obj.migration_plan_dir / "environment.yaml"
        try:
            with open(environment_file, "r") as f:
                obj.environment = yaml.load(f)
        except FileNotFoundError:
            pass

        obj.storage_gateway = StorageGateway.load(obj.migration_plan_dir, obj_version)
        obj.mapped_collection = MappedCollection.load(
            obj.migration_plan_dir, obj_version
        )

        gcdir = obj.migration_plan_dir / "guest_collections"

        guest_collections = []
        if os.path.exists(gcdir):
            for gcid in os.listdir(gcdir):
                gc = GuestCollection.load(gcdir, gcid, obj_version)
                if gc:
                    guest_collections.append(gc)

        obj.guest_collections = guest_collections
        obj.gsi_authz_configuration = GsiAuthzConfiguration.load(
            obj.migration_plan_dir, obj_version
        )
        credentials: List[Credential] = []
        if packaging.version.parse(obj_version) < packaging.version.parse(
            VERSION_INDEPENDENT_CREDENTIAL
        ):
            # Older plans had credentials associated with specific guest
            # collections
            for gc in obj.guest_collections:
                if gc.credential:
                    credentials.append(gc.credential)
        else:
            cred_dir = obj.migration_plan_dir / "credentials"
            if os.path.exists(cred_dir):
                for cred_id_yaml in os.listdir(cred_dir):
                    cred = Credential.load(cred_dir, cred_id_yaml, obj_version)
                    if cred:
                        credentials.append(cred)
                        for gc in obj.guest_collections:
                            if gc.credential_id == cred.id:
                                gc.credential = cred
        obj.credentials = credentials

        return obj

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MigrationPlan):
            return NotImplemented

        return self.__getstate__() == other.__getstate__()

    def __getstate__(self) -> MigrationPlanDict:
        return {
            "gsi_authz_configuration": self.gsi_authz_configuration,
            "storage_gateway": self.storage_gateway,
            "mapped_collection": self.mapped_collection,
            "guest_collections": self.guest_collections,
            "credentials": self.credentials,
            "environment": self.environment if self.environment else "default",
            "version": self.version,
        }

    def __setstate__(self, state: MigrationPlanDict) -> Any:
        self.gsi_authz_configuration = state["gsi_authz_configuration"]
        self.storage_gateway = state["storage_gateway"]
        self.mapped_collection = state["mapped_collection"]
        self.guest_collections = state["guest_collections"]
        self.credentials = state["credentials"]
        self.environment = state["environment"]
        self.version = state.get("version", __version__)
