import enum
import functools
import re
from typing import (
    cast,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    overload,
    Type,
    TypeVar,
)
from typing_extensions import ParamSpec, Protocol

import base62
from flask import current_app, g, request
from werkzeug.exceptions import HTTPException
from werkzeug.routing import BaseConverter

import marshmallow as ma
import globus.manager.api.errors as errors
from globus.manager.api.utils.server_timing import ServerTiming
from .openapi import (
    ensure_openapi,
    AuthorizerExtension,
    OpenAPIAnnotated,
    PublicAuthorizer,
    RoleAuthorizer,
    StorageGatewayAuthorizer,
    UserCredentialOwnerAuthorizer,
)
from globus.manager.api.authentication import authenticated as auth_authenticated
from globus.manager.api.authorization import (
    requires_one_role as auth_requires_one_role,
    requires_credential_owner as auth_requires_credential_owner,
    requires_subscription as auth_requires_subscription,
    requires_endpoint_visibility as auth_requires_endpoint_visibility,
)
from globus.manager.constants import RoleType


T = TypeVar("T")
P = ParamSpec("P")


@overload
def expects_form(fn: Callable[P, T]) -> OpenAPIAnnotated[P, T]:
    ...


@overload
def expects_form(
    *,
    schema: Optional[ma.Schema],
    **metadata: Any,
) -> Callable[[Callable[P, T]], OpenAPIAnnotated[P, T]]:
    ...


def expects_form(
    fn: Callable[P, T] = None, *, schema: Optional[ma.Schema] = None, **metadata
):
    return _expects_request_body(
        fn, media_type="application/x-www-form-urlencoded", schema=schema, **metadata
    )


@overload
def expects_json(fn: Callable[P, T]) -> OpenAPIAnnotated[P, T]:
    ...


@overload
def expects_json(
    *,
    schema: Optional[ma.Schema],
    **metadata: Any,
) -> Callable[[Callable[P, T]], OpenAPIAnnotated[P, T]]:
    ...


def expects_json(
    fn: Callable[P, T] = None, *, schema: Optional[ma.Schema] = None, **metadata
):
    return _expects_request_body(
        fn, media_type="application/json", schema=schema, **metadata
    )


def _expects_request_body(
    fn: Callable[P, T] = None,
    *,
    media_type: str,
    schema: Optional[ma.Schema] = None,
    **metadata: Any,
):
    """
    Decorate a route to require an input body and validate against a
    marshmallow schema.

    :param fn:
        View method to execute if the expectation is met. When called, this
        decorator will populate `g.validated_data`. If there is no schema
        associated with this decorator, the value of `g.validated_data` is
        set to `None`, otherwise it is the output of deserializing the input
        object with the marshmallow schema.
    :param media_type:
        Expected content type for this message body.
    :param schema:
        A schema instance to use to validate the input.
    :param metadata:
        Optional OpenAPI metadata for this request.

    :raises UnsupportedMediaError:
        The input content-type does not match media_type
    :raises BadRequestError:
        The input could not be loaded from the HTTP request.
    :raises ValidationError:
        The input does not validate against the schema.
    """
    if fn is None:
        return functools.partial(
            _expects_request_body, media_type=media_type, schema=schema, **metadata
        )

    fn = ensure_openapi(fn)
    if "description" in metadata:
        fn.metadata.request_body["description"] = metadata["description"]

    if metadata.get("required", False):
        fn.metadata.request_body["required"] = True

    schema_ref = f"#/components/schemas/{schema.__class__.__name__}"
    if schema_ref.endswith("Private"):
        schema_ref = schema_ref[: -len("Private")]
    if schema_ref.endswith("Patch"):
        schema_ref = schema_ref[: -len("Patch")]
    fn.metadata.request_body["content"] = {media_type: {"schema": {"$ref": schema_ref}}}

    fn = returns_error(
        error_type=errors.UnsupportedMediaError,
        response_id="UnsupportedMediaError",
    )(fn)

    fn = returns_error(
        error_type=errors.BadRequestError, response_id="BadRequestError"
    )(fn)

    fn = returns_error(
        error_type=errors.ValidationError,
        error_detail=[errors.InvalidInput],
        response_id="ValidationError",
    )(fn)

    @functools.wraps(fn)
    def decorated(*args: P.args, **kwargs: P.kwargs) -> T:
        if media_type == "application/json":
            if not request.is_json:
                raise errors.UnsupportedMediaError()

            try:
                data = request.get_json()
            except HTTPException as ex:
                raise errors.BadRequestError(ex.description)
        elif media_type == "application/x-www-form-urlencoded":
            try:
                data = request.form
            except HTTPException as ex:
                raise errors.BadRequestError(ex.description)

        g.validated_data = None

        if schema:
            try:
                g.validated_data = schema.load(data)
                current_app.logger.debug(
                    "Validated Data ({}): {}".format(
                        schema.__class__.__name__ if schema else "any",
                        g.validated_data,
                    )
                )
            except ma.ValidationError as e:
                current_app.logger.error(
                    "deserialization error: {} {} using schema {}".format(
                        e, data, schema
                    )
                )
                raise errors.ValidationError(exc=e)

        assert fn is not None
        return fn(*args, **kwargs)

    return decorated


# Server metrics we want to provide to API consumers
db_read = ("db-read", "Database Reads")
db_write = ("db-write", "Database Writes")
transfer_api = ("t-api", "Transfer API")
auth_api = ("auth-api", "Auth API")
letsencrypt = ("letsencrypt", "LetsEncrypt")


def server_timing(fn: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        g.server_timing = ServerTiming()

        assert fn is not None
        return fn(*args, **kwargs)

    return decorated_function


class Base62Marker:
    """
    Generic safe string marker implementation. This is used as the
    default marker class for the paginated decorator.
    """

    def __init__(self, marker: str = None, marker_data: str = None) -> None:
        """
        Initializer for the Base62Marker class. One of the two named parameters
        can be used to initialize the marker.

        :param marker:
            The base62 encoded representation of this marker. This is typically
            used when parsing the marker query parameter to a route.
        :param marker_data:
            A string version of the data which will be stored in the marker.
        """
        if marker:
            self.decode(marker)
            if marker_data:
                raise ValueError("Invalid marker")
        elif marker_data:
            self.marker = marker_data
        else:
            raise ValueError("Missing required marker or marker_data")

    @property
    def marker(self) -> str:
        """
        The string version of the data stored in the marker.
        """
        return self._marker

    @marker.setter
    def marker(self, value: str) -> None:
        self._marker = value

    def decode(self, marker: str) -> None:
        """
        Decode the base62 marker string and store its value in the
        `marker` property
        """
        self.marker = base62.decodebytes(marker).decode("utf-8")

    def encode(self) -> str:
        """
        Encode and return the `marker` property as a base62 string.
        """
        return base62.encodebytes(self.marker.encode("utf-8"))

    def __str__(self) -> str:
        return self.marker


class Paginator(Protocol):
    def __init__(self, marker: str) -> None:
        ...

    @property
    def marker(self) -> str:
        ...

    @marker.setter
    def marker(self, str) -> None:
        ...

    def decode(self, marker: str) -> None:
        ...

    def encode(self) -> str:
        ...


@overload
def paginated(fn: Callable[P, T]) -> Callable[P, T]:
    ...


@overload
def paginated(
    *, marker_class: Type[Paginator] = None
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...


def paginated(fn: Callable[P, T] = None, *, marker_class: Type[Paginator] = None):
    """
    Decorate a route to parse and validate pagination parameters `marker` and
    `page_size`. The results of parsing are stored in `g.marker` and
    `g.page_size`.

    :param fn:
        The route handler to call after parsing the optional pagination
        parameters.
    :param marker_class:
        An optional class that implements a constructor that can be passed a
        `marker` named parameter with encode() and decode() methods to parse
        and serialize the marker. If not present or None, the
        :py:class:`globus.manager.api.blueprints.decorators.Base62Marker` class
        is used.
    """
    if fn is None:
        return functools.partial(paginated, marker_class=marker_class)

    if marker_class is None:
        marker_class = Base62Marker

    @query_parameter(
        name_spec="marker",
        description="Pagination marker for a paginated response",
        title="marker_query_parameter",
    )
    @query_parameter(
        name_spec="<int:page_size>",
        description="Maximum page size for a paginated response",
        title="page_size_query_parameter",
        default=16,
        schema={
            "type": "integer",
            "maximum": 100,
        },
    )
    @functools.wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        page_size = cast(int, kwargs["page_size"])
        marker_string = cast(str, kwargs["marker"])
        marker = None

        if page_size < 1 or page_size > 100:
            raise errors.BadRequestError(
                "The page_size parameter must be a positive value between 1 and 100"
            )

        if marker_string:
            try:
                assert marker_class is not None
                marker = marker_class(marker=marker_string)
            except (ValueError, AttributeError, UnicodeDecodeError):
                raise errors.BadRequestError("Invalid marker")

        kwargs["marker"] = marker
        assert fn is not None
        return fn(*args, **kwargs)

    return decorated_function


E = TypeVar("E", bound=enum.Enum)


class ConverterWithMetadata(BaseConverter):
    metadata: Dict[str, Any]


def enum_list_converter(enum_type: Type[E]) -> Type[ConverterWithMetadata]:
    class _EnumListConverter(ConverterWithMetadata):
        metadata = {
            "style": "form",
            "explode": False,
            "schema": {
                "type": "array",
                "items": {"type": "string", "enum": [e.name for e in enum_type]},
            },
        }

        def to_python(self, value: str) -> List[E]:
            if value:
                return [enum_type(v) for v in value.split(",")]
            return []

        def to_url(self, value: Any) -> str:
            if isinstance(value, enum_type):
                return value.name
            else:
                return str(value)

    return _EnumListConverter


@overload
def query_parameter(fn: Callable[P, T]) -> Callable[P, T]:
    ...


@overload
def query_parameter(
    *,
    name_spec: str,
    converter: Optional[Type[ConverterWithMetadata]] = None,
    **metadata,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...


def query_parameter(
    fn: Callable[P, T] = None,
    *,
    name_spec: str = None,
    converter: Optional[Type[ConverterWithMetadata]] = None,
    **metadata: Any,
):
    """
    Decorate a route to parse and validate a query parameter using the
    same converters as flask uses for query parameters. The converted value
    is passed to the route using the name of parameter.

    :param fn:
        The route handler to call after parsing the optional pagination
        parameters.
    :param name_spec:
        A string containing the name of the query parameter or a string
        of the form <converter:name>, in which case the flask URL
        map converter will be invoked.
    :param converter":
        Optional subclass of BaseConverter to apply to the value, if not
        registered with the current_app's url_map.
    :param **metadata:
        OpenAPI metadata for this query parameter. Currently "default", and
        "required" are supported. If required is True, and the parameter is omitted,
        a ValueError() exception is raised.
    """
    if fn is None:
        return functools.partial(
            query_parameter,
            name_spec=name_spec,
            converter=converter,
            **metadata,
        )

    converter_class = converter
    assert name_spec is not None
    match = re.match("<(.*):(.*)>", name_spec)
    converter_name = None
    if match:
        converter_name = match.group(1)
        name = match.group(2)
    else:
        name = name_spec

    fn = ensure_openapi(fn)
    parameter: Dict[str, Any] = {"name": name, "in": "query", **metadata}
    if "schema" not in parameter:
        parameter["schema"] = {"type": "string"}
    if "default" in parameter:
        parameter["schema"]["default"] = parameter.pop("default")
    if converter_name:
        parameter["schema"]["format"] = converter_name

    if converter:
        parameter.update(converter.metadata)
    fn.metadata.parameters.append(parameter)
    required = cast(bool, metadata.get("required", False))

    if required:
        fn = returns_error(
            error_type=errors.BadRequestError, response_id="BadRequestError"
        )(fn)

    @functools.wraps(fn)
    def decorated_function(*args: P.args, **kwargs: P.kwargs) -> T:
        cv: Optional[BaseConverter] = None
        if converter_name:
            cv = current_app.url_map.converters.get(converter_name)(current_app.url_map)
        else:
            if converter_class:
                cv = converter_class(current_app.url_map)

        raw_value = request.args.get(name)
        default = metadata.get("default", None)
        if raw_value is None and required:
            raise errors.BadRequestError(
                f"Missing required value for query parameter {name}"
            )
        if cv and raw_value is not None:
            try:
                kwargs[name] = cv.to_python(raw_value)
            except Exception:
                raise errors.BadRequestError(
                    f"Invalid value for query parameter {name}"
                )

        elif raw_value is not None:
            kwargs[name] = raw_value
        else:
            kwargs[name] = default

        assert fn is not None
        return fn(*args, **kwargs)

    return decorated_function


@overload
def result_data(fn: Callable[P, T]) -> OpenAPIAnnotated[P, T]:
    ...


@overload
def result_data(
    *,
    schemas: List[ma.Schema],
    max_items: Optional[int] = None,
    http_status: Optional[int] = 200,
    **metadata: Any,
) -> Callable[[Callable[P, T]], OpenAPIAnnotated[P, T]]:
    ...


def result_data(
    fn: Callable[P, T] = None,
    *,
    schemas: Optional[List[ma.Schema]] = None,
    max_items: Optional[int] = None,
    http_status: Optional[int] = 200,
    **metadata: Any,
):
    if fn is None:
        return functools.partial(
            result_data,
            max_items=max_items,
            schemas=schemas,
            http_status=http_status,
            **metadata,
        )
    fn = ensure_openapi(fn)

    def schema_ref(schema: ma.Schema) -> Dict[str, str]:
        ref = f"#/components/schemas/{schema.__class__.__name__}"
        return {"$ref": ref}

    if schemas is None:
        schemas = []

    response = fn.metadata.responses[str(http_status)] = {
        "description": metadata.get("description", "Operation response"),
        "content": {
            "application/json": {
                "schema": {
                    "allOf": [
                        {
                            "type": "object",
                            "properties": {
                                "code": {
                                    "type": "string",
                                    "enum": [f"{metadata.get('code', 'success')}"],
                                },
                                "data": {
                                    "type": "array",
                                    "items": {
                                        "oneOf": [
                                            schema_ref(schema) for schema in schemas
                                        ]
                                    },
                                },
                            },
                        },
                        {"$ref": "#/components/schemas/Result"},
                    ]
                }
            }
        },
    }
    schema = response["content"]["application/json"]["schema"]
    if not schemas:
        schema["$ref"] = schema["allOf"][1]["$ref"]
        schema.pop("allOf")
    elif max_items:
        specialization = schema["allOf"][0]
        specialization["properties"]["data"]["maxItems"] = max_items

    return fn


@overload
def returns_error(fn: Callable[P, T]) -> OpenAPIAnnotated[P, T]:
    ...


@overload
def returns_error(
    *,
    error_type: Type[errors.APIBaseError],
    error_detail: Optional[List[Type[errors.ErrorDetail]]] = None,
    response_id: Optional[str] = None,
) -> Callable[[Callable[P, T]], OpenAPIAnnotated[P, T]]:
    ...


def returns_error(
    fn: Callable[P, T] = None,
    *,
    error_type: Type[errors.APIBaseError] = None,
    error_detail: Optional[List[Type[errors.ErrorDetail]]] = None,
    response_id: Optional[str] = None,
):
    if fn is None:
        return functools.partial(
            returns_error,
            error_type=error_type,
            error_detail=error_detail,
            response_id=response_id,
        )
    assert error_type is not None
    fn = ensure_openapi(fn)
    fn.metadata.returns_error(
        error_type=error_type, error_detail=error_detail, response_id=response_id
    )

    return fn


@overload
def responds_with_redirect(fn: Callable[P, T]) -> OpenAPIAnnotated[P, T]:
    ...


@overload
def responds_with_redirect(
    *, http_status: int, description: str = None
) -> Callable[[Callable[P, T]], OpenAPIAnnotated[P, T]]:
    ...


def responds_with_redirect(
    fn: Callable[P, T] = None, http_status: int = 302, description: str = None
):
    if fn is None:
        return functools.partial(
            responds_with_redirect,
            http_status=http_status,
            description=description,
        )

    fn = ensure_openapi(fn)

    status_description = {
        302: "Found",
        303: "See Other",
    }[http_status]

    if str(http_status) not in fn.metadata.responses:
        fn.metadata.responses[str(http_status)] = {
            "description": description or status_description,
        }

    return fn


def authenticated(fn: Callable[P, T]) -> OpenAPIAnnotated[P, T]:
    fn = returns_error(
        error_type=errors.UnauthorizedError,
        error_detail=[errors.MissingRequiredScopes],
        response_id="Unauthorized",
    )(fn)

    return auth_authenticated(fn)


def requires_one_role(fn: Callable[P, T] = None, **kwargs):
    if fn is None:
        return functools.partial(requires_one_role, **kwargs)

    fn = auth_requires_one_role(**kwargs)(fn)
    assert fn is not None
    fn = returns_error(
        error_type=errors.ForbiddenError,
        error_detail=[errors.MissingRequiredRole],
    )(fn)
    role_names: List[str] = kwargs["role_names"]
    fn.metadata.authorization.append(RoleAuthorizer(roles=role_names))

    return fn


def requires_credential_owner(fn: Callable[P, T] = None, **kwargs):
    if fn is None:
        return functools.partial(requires_credential_owner, **kwargs)

    fn = returns_error(
        error_type=errors.ForbiddenError,
        error_detail=[errors.NotResourceOwner],
    )(fn)
    fn = document_authorization(
        authorization=UserCredentialOwnerAuthorizer(scopes=["{user_credential_id}"])
    )(fn)

    return cast(OpenAPIAnnotated[P, T], auth_requires_credential_owner(**kwargs)(fn))


def requires_subscription(fn: Callable[P, T] = None, **kwargs):
    if fn is None:
        return functools.partial(requires_subscription, **kwargs)

    fn = returns_error(
        error_type=errors.SubscriptionRequiredError,
        error_detail=[errors.SubscriptionRequired],
    )(fn)

    return cast(OpenAPIAnnotated[P, T], auth_requires_subscription(**kwargs)(fn))


def requires_endpoint_visibility(fn: Callable[P, T]) -> OpenAPIAnnotated[P, T]:
    fn = auth_requires_endpoint_visibility(fn)

    fn = ensure_openapi(fn)
    fn.metadata.authorization.append(PublicAuthorizer())
    fn.metadata.authorization.append(
        RoleAuthorizer(
            roles=[
                RoleType.endpoint_owner.value,
                RoleType.endpoint_administrator.value,
                RoleType.endpoint_activity_manager.value,
                RoleType.endpoint_activity_monitor.value,
                RoleType.collection_administrator.value + ":*",
                RoleType.collection_activity_manager.value + ":*",
                RoleType.collection_activity_monitor.value + ":*",
                RoleType.collection_access_manager.value + ":*",
            ]
        )
    )
    fn.metadata.authorization.append(StorageGatewayAuthorizer(scopes=["*"]))

    return fn


@overload
def document_authorization(
    *,
    authorization: AuthorizerExtension,
) -> Callable[[Callable[P, T]], OpenAPIAnnotated[P, T]]:
    ...


@overload
def document_authorization(fn: Callable[P, T]) -> OpenAPIAnnotated[P, T]:
    ...


def document_authorization(
    fn: Callable[P, T] = None,
    *,
    authorization: AuthorizerExtension = None,
):
    if fn is None:
        return functools.partial(document_authorization, authorization=authorization)
    fn = ensure_openapi(fn)
    if authorization:
        fn.metadata.authorization.append(authorization)

    return fn
