import enum
from typing import List, Optional
import uuid

from flask import (
    current_app,
    g,
    jsonify,
    request,
    Response,
)
from globus.manager.api.blueprints.decorators import (
    authenticated,
    Base62Marker,
    enum_list_converter,
    expects_json,
    paginated,
    Paginator,
    query_parameter,
    requires_one_role,
    requires_subscription,
    result_data,
    returns_error,
)
from globus.manager.api.blueprints.utils import (
    create_object,
    delete_object,
)
import globus.manager.api.errors as errors
from globus.manager.api.models import (
    Collection,
    ConfigurationObject,
    Endpoint,
    Role,
    Result,
)
from globus.manager.api.schemas.api.result import result_schema
from globus.manager.api.schemas.api.role import role_schema
from globus.manager.constants import RoleType
from .openapi import OpenAPIBlueprint


blueprint = OpenAPIBlueprint(
    "roles",
    __name__,
    tag="Roles",
    description="""
    The client identity that created an Endpoint has full control over
    management aspects of the endpoint, including the properties, collections,
    the ACL, and role assignments.  Other roles enable other identities to act
    on various subsets of endpoint and collection management capabilities.  See
    the
    [endpoint roles](https://docs.globus.org/globus-connect-server/v5.4/reference/endpoint/role/)
    and
    [collection roles](https://docs.globus.org/globus-connect-server/v5.4/reference/collection/role/)
    reference pages for information about the roles.  All API endpoints
    below indicate which roles are required to perform the given operation.

    """,
    menu_weight=4,
    url_prefix="/api/",
)


def get_collection_from_collection_id_query() -> Optional[Collection]:
    """
    Resolves the collection_id query parameter to a Collection
    model for authorization
    """
    collection_id = request.args.get("collection_id")
    collection: Optional[Collection] = None

    if collection_id:
        collection = Collection.query.filter(
            ~Collection.is_deleted,  # type: ignore
            Collection.id == collection_id,
        ).one_or_none()
        if collection:
            g.collection = collection

    return collection


def get_collection_from_role_object() -> Optional[Collection]:
    """
    Resolves the collection_id from the parsed input object
    to a Collection model for authorization
    """
    collection_id = g.validated_data.get("collection")
    if collection_id:
        collection = Collection.query.filter(
            ~Collection.is_deleted,  # type: ignore
            Collection.id == str(collection_id),
        ).one_or_none()

        if not collection:
            raise errors.ForbiddenError()
        g.collection = collection
        return g.collection
    return None


def get_collection_from_role_id_path() -> Optional[Collection]:
    """
    Resolves the collection_id to the Collection if the
    path parameter role_id resolves to a Role assignment for
    a collection.
    """
    role_id = request.view_args.get("role_id")
    if role_id:
        role_id = str(role_id)
        role = Role.query.filter(
            ~Role.is_deleted,  # type: ignore
            Role.id == role_id,
        ).one_or_none()
        if role and role.resource:
            return role.collection
    return None


class ListRolesInclude(enum.Enum):
    all_roles = "all_roles"


@blueprint.route(
    "/roles",
    methods=["GET"],
    summary="List roles",
    operationId="listRoles",
)
@result_data(
    description="List roles response",
    schemas=[role_schema],
)
@authenticated
@query_parameter(
    name_spec="include",
    description="""
    Flag indicating whether to request all roles assignments for the endpoint
    or collection.
    """,
    converter=enum_list_converter(ListRolesInclude),
    default=[],
)
@query_parameter(
    name_spec="<uuid:collection_id>",
    description="ID of the collection",
    title="collection_id_query_parameter",
)
@paginated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_from_collection_id_query,
)
def roles_list(
    *,
    collection_id: Optional[uuid.UUID],
    include: List[ListRolesInclude],
    page_size: int,
    marker: Optional[Paginator]
) -> Response:
    """
    Get the endpoint's or a collection's list of role associations.

    If the `collection_id` query parameter is passed to this operation, then
    the roles related to that collection are returned. Otherwise, this
    operation returns endpoint roles.

    The `include` parameter determines whether this operation returns all roles
    relevant to the resource or only those that the caller has.

    To obtain information about *all* roles, the caller must pass the
    "all_roles" value as the value of the "include" parameter.  This requires
    the "administrator" role for the endpoint Or Collection the role is
    associated with.
    """
    include_all_roles = ListRolesInclude.all_roles in include

    if collection_id is not None:
        role_query = Role.query.filter(
            ~Role.is_deleted,  # type: ignore
            Role.is_collection_role,
            Role.resource == str(collection_id),
        )
    else:
        role_query = Role.query.filter(
            ~Role.is_deleted,  # type: ignore
            Role.is_endpoint_role,
        )

    # TODO: Handle group role assignments
    if not include_all_roles:
        role_query = role_query.filter(
            Role.identity_id.in_([i["id"] for i in g.auth_state.identity_set_data])  # type: ignore
        )

    if marker:
        role_query = role_query.filter(ConfigurationObject.id >= marker.marker)

    if not current_app.subscription:
        role_query = role_query.filter(Role.role == RoleType.endpoint_owner)

    role_query = role_query.order_by(ConfigurationObject.id).limit(page_size + 1)

    roles = []

    for role in role_query:
        roles.append(role_schema.dump(role))

    next_page_token = None  # type: Optional[Base62Marker]
    if len(roles) > page_size:
        next_page_token = Base62Marker(marker_data=roles.pop()["id"])

    res = result_schema.dump(Result(data=roles, marker=next_page_token))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route("/roles", methods=["POST"], summary="Create a role")
@result_data(
    description="Create role response",
    schemas=[role_schema],
    operationId="postRole",
)
@authenticated
@expects_json(
    schema=role_schema, description="New role assignment document", required=True
)
@requires_subscription
@requires_one_role(
    role_names=["endpoint:administrator", "collection:administrator:{collection_id}"],
    require_session=True,
    collection_selector=get_collection_from_role_object,
)
@returns_error(
    error_type=errors.BadRequestError,
)
@returns_error(
    error_type=errors.ResourceExistsError,
)
def roles_post() -> Response:
    """
    Assign a role to an identity or group for the endpoint or a specified
    collection based on the `collection_id` query parameter.

    On success returns a copy of the created role with the system generated id
    added.
    """
    # Ignore input id
    g.validated_data.pop("id", None)

    input_role = g.validated_data
    role_collection_uuid = input_role.get("collection")
    role_collection = None  # type: Optional[str]
    if role_collection_uuid:
        role_collection = str(role_collection_uuid)
        if g.collection.is_mapped_collection and input_role["role"] == "access_manager":
            raise errors.BadRequestError("Invalid role for mapped collection")
    else:
        role_collection = None
        if input_role["role"] == "access_manager":
            raise errors.BadRequestError("Invalid role for endpoint")

    if input_role["role"] == "owner":
        raise errors.BadRequestError("Owner role may not be created explicitly")

    # Validate the principal if it is a identity. Auth does not have a proper
    # groups API for validating groups; you can only validate groups you belong
    # to.
    principal = input_role.get("principal")
    identity_urn_prefix = "urn:globus:auth:identity:"
    if principal.startswith(identity_urn_prefix):
        auth_identity = current_app.auth_client.get_identities(
            ids=[principal[len(identity_urn_prefix) :]]
        ).data.get("identities", [])
        if len(auth_identity) != 1 or auth_identity[0]["status"] not in [
            "unused",
            "used",
            "private",
        ]:
            raise errors.BadRequestError("Role principal is not a valid Globus account")
        if auth_identity[0]["identity_type"] == "email":
            raise errors.BadRequestError(
                "Role principal must be an account with a non-email identity_type"
            )

    with current_app.appsync_client.pause_notifications():
        if Role.query.filter(
            ~Role.is_deleted,  # type: ignore
            Role.principal == input_role["principal"],
            Role.role_name == input_role["role"],
            Role.resource.is_(role_collection),  # type: ignore
        ).one_or_none():
            raise errors.ResourceExistsError()

        try:
            endpoint = Endpoint.query.one()
            role = Role(doc=input_role, endpoint=endpoint)
        except errors.APIBaseError:
            raise
        except Exception as e:
            current_app.logger.info("Unable to create role: %s", str(e))
            raise errors.InternalServerError("Unable to create role")

        create_object(role)

        res = result_schema.dump(
            Result(
                message="Created new role {}".format(role.id),
                data=[role_schema.dump(role)],
            )
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.created_objects.append(role.id)

    return response


@blueprint.route(
    "/roles/<uuid:role_id>",
    methods=["DELETE"],
    summary="Delete a role",
    path_parameters={"role_id": "Id of the role"},
    operationId="deleteRole",
)
@result_data(
    description="Delete role response",
    schemas=[],
)
@authenticated
@requires_subscription
@requires_one_role(
    role_names=["endpoint:administrator", "collection:administrator:{collection_id}"],
    collection_selector=get_collection_from_role_id_path,
)
@returns_error(
    error_type=errors.ConflictError,
)
def roles_delete(role_id: uuid.UUID) -> Response:
    """
    Delete one of the roles on the endpoint.
    """

    current_app.logger.debug("DELETE role {}".format(role_id))
    role_id_str = str(role_id)

    with current_app.appsync_client.pause_notifications():
        role = Role.query.filter(
            ~Role.is_deleted,  # type: ignore
            Role.id == role_id_str,
        ).one_or_none()

        if not role:
            raise errors.NotFoundError(
                error_description="Role {} not found".format(role_id)
            )

        if role.primary_role:
            raise errors.ConflictError("Unable to delete primary role")

        delete_object(role)

        res = result_schema.dump(Result(message="Deleted role {}".format(role_id)))

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.deleted_objects.append(role.id)

    return response


@blueprint.route(
    "/roles/<uuid:role_id>",
    methods=["GET"],
    summary="Get a role",
    path_parameters={"role_id": "Id of the role"},
    operationId="getRole",
)
@result_data(
    description="Get role response",
    schemas=[role_schema],
    max_items=1,
)
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_from_role_id_path,
)
def roles_get_by_id(role_id: uuid.UUID) -> Response:
    """
    Get one of the role assignments on this endpoint.
    """

    role_id_str = str(role_id)
    role = Role.query.filter(
        ~Role.is_deleted,  # type: ignore
        Role.id == role_id_str,
    ).one_or_none()
    if not role:
        raise errors.NotFoundError()

    res = result_schema.dump(Result(data=[role_schema.dump(role)]))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response
