from typing import Any, Dict, Optional
from typing_extensions import Protocol
import urllib.parse
import uuid

import jwt

from flask import (
    Flask,
    current_app,
    g,
    jsonify,
    make_response,
    redirect,
    render_template,
    request,
    url_for,
)
from oauth2client import client as oauth_client
from globus.manager.api.models import (
    db,
    Endpoint,
    Result,
    StorageGateway,
    UserCredential,
)
from globus.manager.api.blueprints.decorators import (
    authenticated,
    expects_form,
    query_parameter,
    result_data,
    responds_with_redirect,
)
from globus.manager.api.authorization import StorageGatewayAuthorizer
from globus.manager.utils.update_query_string import update_query_string
import globus.manager.api.errors as errors
from globus.manager.api.blueprints.openapi import OpenAPIBlueprint
from globus.manager.api.blueprints.utils import (
    create_object,
    update_object,
)
from globus.manager.api.schemas.api_v1.request import user_credential_form_post_schema
from globus.manager.api.schemas.api import result_schema
from sqlalchemy.orm.exc import NoResultFound


class OAuthCredentialConnector(Protocol):
    DISPLAY_NAME: str

    def initiate_flow(
        self,
        storage_gateway: StorageGateway,
        authorized_identity: Dict[str, Any],
        **kwargs,
    ) -> oauth_client.OAuth2WebServerFlow:
        ...


class OAuthCredentialAPI:
    __instance: Optional["OAuthCredentialAPI"] = None
    __blueprint_name: str = "oauth_credential_api"
    __connectors: Dict[str, OAuthCredentialConnector] = {}
    blueprint: OpenAPIBlueprint

    def __new__(cls, app: Flask):
        if cls.__instance is None:
            inst = super().__new__(cls)
            inst.blueprint = OpenAPIBlueprint(
                cls.__blueprint_name,
                __name__,
                url_prefix="/api/v1",
                tag="OAuth Credential API",
                description="""
                The Globus Connect Server OAuth Credential API provides routes
                for managing credentials for Storage Gateways that depend on a
                browser-based exchange to obtain credentials to access a
                storage system.
                """,
                schema_name="oauth-credential-api.yaml",
            )
            inst.blueprint.add_url_rule(
                "/user_credentials",
                "user_credentials_post",
                inst.user_credentials_post,
                methods=["POST"],
                summary="Create a user credential",
            )
            # Backward compatibility name
            inst.blueprint.add_url_rule(
                "/authcallback_google",
                "authcallback_google",
                inst.authcallback,
                methods=["GET"],
                summary="OAuth callback",
                deprecated=True,
            )
            inst.blueprint.add_url_rule(
                "/authcallback",
                "authcallback",
                inst.authcallback,
                methods=["GET"],
                summary="OAuth callback",
            )
            inst.blueprint.add_url_rule(
                "/authclicomplete",
                "authclicomplete",
                inst.authclicomplete,
                methods=["GET"],
                summary="OAuth callback for CLI apps",
            )
            inst.blueprint.register_error_handler(
                errors.APIBaseError, inst.handle_validation_error
            )
            cls.__instance = inst
        if cls.__instance.blueprint.name not in app.blueprints.keys():
            app.register_blueprint(cls.__instance.blueprint)
        return cls.__instance

    @classmethod
    def add_connector(
        cls, connector_id: str, connector: OAuthCredentialConnector
    ) -> None:
        cls.__connectors[connector_id] = connector

    @expects_form(
        schema=user_credential_form_post_schema, description="User credential request"
    )
    @responds_with_redirect(http_status=303, description="Redirection to OAuth flow")
    @authenticated
    def user_credentials_post(self):
        """
        Create a user credential on a Storage Gateway. This
        initiates a flow to an OAuth service
        to authenticate the user and generate an authentication
        token to allow access to a storage system.
        """
        post_data = g.validated_data

        # confirm post data is for a valid storage gateway
        storage_gateway = None
        try:
            storage_gateway = StorageGateway.query.filter_by(
                deleted=False, id=post_data["storage_gateway"]
            ).one()
        except NoResultFound:
            current_app.logger.info(
                "Attempt to create a user credential with an unknown or "
                "disabled storage gateway: {}".format(post_data["storage_gateway"])
            )
            raise errors.BadRequestError(
                "storage_gateway does not exist, is "
                "disabled, or does not allow guest collections"
            )

        connector = self.__connectors.get(storage_gateway.connector_id)
        if not connector:
            raise errors.UnsupportedMediaError()

        authorizer = StorageGatewayAuthorizer(storage_gateway)
        try:
            authorized_identity = authorizer.authorize_identity(
                post_data["identity_id"], g.auth_state
            )
        except errors.ForbiddenError as e:
            query_updates = {
                "error": e.error,
                "error_description": e.error_description,
            }
            if e.detail:
                if isinstance(e.detail, errors.NotFromAllowedDomain):
                    query_updates["domains"] = ",".join(e.detail.allowed_domains)
                elif isinstance(e.detail, errors.AuthenticationTimeout):
                    query_updates["identities"] = ",".join(e.detail.identities)
            if "redirect_uri" in post_data:
                redirect_to = update_query_string(
                    post_data["redirect_uri"], query_updates
                )
                return redirect(redirect_to)
            else:
                raise

        post_data["authorized_id_data"] = authorized_identity

        app_secret = Endpoint.query.one().oauth_state_key
        state = jwt.encode(post_data, app_secret, algorithm="HS256")

        flow = connector.initiate_flow(
            storage_gateway,
            authorized_identity,
            state=state,
            response_type="code",
        )

        return redirect(flow.step1_get_authorize_url(), code=303)

    @query_parameter(
        name_spec="<uuid:user_credential_id>",
        description="Unique identifier of the new credential",
    )
    @result_data(schemas=[], description="Authentication callback response")
    def authclicomplete(self, user_credential_id: uuid.UUID):
        """
        Landing page for auth flow completion, suitable for use with a
        non-web application that uses a browser to complete the flow.
        """
        if user_credential_id:
            if "text/html" == request.accept_mimetypes.best:
                response = make_response(
                    render_template(
                        "authclicomplete.jinja2",
                        user_credential_id=str(user_credential_id),
                    )
                )
                return response
            else:
                return jsonify(
                    result_schema.dump(
                        Result(message=f"Created credential {user_credential_id}")
                    )
                )
        else:
            raise errors.BadRequestError("Credential creation failed")

    @query_parameter(
        name_spec="state", description="Encrypted authorization context", required=True
    )
    @query_parameter(
        name_spec="error",
        description="Error information from the OAuth provider",
    )
    @query_parameter(name_spec="code", description="OAuth code response")
    @responds_with_redirect(
        http_status=302,
        description="Redirect to page initiating the credential creation",
    )
    def authcallback(
        self, state: str, error: Optional[str] = None, code: Optional[str] = None
    ):
        app_secret = Endpoint.query.one().oauth_state_key

        try:
            state_data = jwt.decode(state, app_secret, algorithms=["HS256"])
        except jwt.InvalidTokenError:
            raise errors.BadRequestError("Invalid state parameter")

        if error:
            # OAuth provider sent us back with an error. Send the user-agent
            # back to the `redirect_uri` that was provided in the original
            # POST to /user_credentials + the error as an added query param.
            redirect_to = update_query_string(
                state_data["redirect_uri"], dict(error=error)
            )

            current_app.logger.error("authcallback error: {}".format(error))

            return redirect(redirect_to)

        try:
            storage_gateway = StorageGateway.query.filter_by(
                deleted=False, id=state_data["storage_gateway"]
            ).one()
        except NoResultFound:
            current_app.logger.error(
                "OAuth callback with an unknown storage_gateway %s.",
                state_data["storage_gateway"],
            )
            raise errors.BadRequestError("storage_gateway does not exist")

        if code:
            connector = self.__connectors.get(storage_gateway.connector_id)
            if not connector:
                current_app.logger.error(
                    "OAuth callback with an unknown connector %s.",
                    storage_gateway.connector_id,
                )
                raise errors.BadRequestError("storage_gateway does not exist")

            # If we have a `code` we can proceed to exchange it for
            # an access token.
            flow = connector.initiate_flow(
                storage_gateway,
                state_data["authorized_id_data"],
                response_type="code",
            )

            try:
                credentials = flow.step2_exchange(code)
            except oauth_client.FlowExchangeError as ex:
                # FlowExchangeError will be the `error` and `error_description`
                # fields concantenated together, so you might see something
                # like `invalid_grantCode was already redeemed`.
                current_app.logger.exception("Error during OAuth authentication flow")
                redirect_to = update_query_string(
                    state_data["redirect_uri"],
                    dict(error="flow_exchange_error", error_description=ex),
                )

                return redirect(redirect_to)

            # if connector doesn't provide an id_token in credentials,
            # use creds to generate one for compatibility
            if not getattr(credentials, "id_token", None) and getattr(
                connector, "id_token", None
            ):
                credentials.id_token = connector.id_token(credentials.access_token)  # type: ignore
            try:
                validated_email = credentials.id_token["email"]
            except (AttributeError, KeyError):
                error_description = (
                    "The {} OAuth endpoint did not provide a valid id token.".format(
                        connector.DISPLAY_NAME,
                    )
                )
                redirect_to = update_query_string(
                    state_data["redirect_uri"],
                    dict(error="invalid_user", error_description=error_description),
                )
                return redirect(redirect_to)

            # let connector decide if there is a better claim for the username
            if getattr(connector, "identity_username", None):
                validated_email = connector.identity_username(  # type: ignore
                    storage_gateway, credentials.id_token
                )
            current_app.logger.debug(
                "Validated email: {}, id_token: {}".format(
                    validated_email, credentials.id_token
                )
            )

            authorized_identity = state_data["authorized_id_data"]
            mapped_to = storage_gateway.map_identity(authorized_identity)

            def _orlist(domains):
                if len(domains) == 1:
                    return "{}".format(domains[0])
                elif len(domains) == 2:
                    return "{} or {}".format(*domains)
                else:
                    return "{}, or {}".format(", ".join(domains[:-1]), domains[-1])

            if not mapped_to:
                error_description = (
                    "Your Globus account {} did not map to a {} account "
                    "usable with this resource. If you have another "
                    "{} Globus account, you can try one of those. "
                    "Otherwise, contact your system administrator.".format(
                        authorized_identity["username"],
                        connector.DISPLAY_NAME,
                        _orlist(storage_gateway.allowed_domains),
                    )
                )
                redirect_to = update_query_string(
                    state_data["redirect_uri"],
                    dict(error="invalid_user", error_description=error_description),
                )
                return redirect(redirect_to)

            if mapped_to.lower() != validated_email.lower():
                error_description = (
                    "Your Globus account mapped to {0}, but you logged in to {1} "
                    "as {2}. If you have a {1} account for {3}, please log in "
                    "to {1} with that, otherwise contact your system "
                    "administrator.".format(
                        mapped_to, connector.DISPLAY_NAME, validated_email, mapped_to
                    )
                )
                redirect_to = update_query_string(
                    state_data["redirect_uri"],
                    dict(error="invalid_user", error_description=error_description),
                )
                return redirect(redirect_to)

            if not storage_gateway.check_user_allowed(mapped_to):
                error_description = (
                    "Your Globus account mapped to {}, but that account is disallowed "
                    "access to this resource by its policy. If you have another "
                    "{} Globus account, you can try one of those. Otherwise, "
                    "contact your system administrator.".format(
                        mapped_to,
                        _orlist(storage_gateway.allowed_domains),
                    )
                )
                redirect_to = update_query_string(
                    state_data["redirect_uri"],
                    dict(error="invalid_user", error_description=error_description),
                )

            user_credential = UserCredential.query.filter_by(
                storage_gateway_id=storage_gateway.id,
                username=mapped_to,
                identity_id=state_data["identity_id"],
            ).one_or_none()

            if user_credential:
                update_object(
                    user_credential,
                    lambda o, credentials: o.put_credentials(credentials),
                    credentials,
                )
            else:
                # create new user credential
                model_class = UserCredential.polymorphic_class(
                    storage_gateway.connector_id
                )

                user_credential = model_class(
                    doc={
                        "DATA_TYPE": "user_credential#1.0.0",
                        "identity_id": state_data["identity_id"],
                        "username": mapped_to,
                        "storage_gateway_id": storage_gateway.id,
                    },
                    storage_gateway=storage_gateway,
                    credentials=credentials,
                )
                create_object(user_credential)
            redirect_to = update_query_string(
                state_data["redirect_uri"], dict(user_credential_id=user_credential.id)
            )

            return redirect(redirect_to)

        # if the request has neither an `error` or a `code` it's a bad request.
        raise errors.BadRequestError()

    @property
    def authcallback_google_url(self):
        callback_url = url_for(
            ".".join([self.blueprint.name, "authcallback_google"]),
            _external=True,
            _scheme="https",
        )
        # In the case of a custom domain, the URL will reflect the original
        # data.globus.org name. Rewrite this URL in this case.
        with db.session.no_autoflush:
            endpoint = Endpoint.query.one()

        if endpoint.has_domain:
            gcs_manager_domain = endpoint.domain.domain_name
            parsed = list(urllib.parse.urlparse(callback_url))
            parsed[1] = gcs_manager_domain
            callback_url = urllib.parse.urlunparse(parsed)

        return callback_url

    @property
    def authcallback_url(self):
        callback_url = url_for(
            ".".join([self.blueprint.name, "authcallback"]),
            _external=True,
            _scheme="https",
        )
        # In the case of a custom domain, the URL will reflect the original
        # data.globus.org name. Rewrite this URL in this case.
        with db.session.no_autoflush:
            endpoint = Endpoint.query.one()

        if endpoint.has_domain:
            gcs_manager_domain = endpoint.domain.domain_name
            parsed = list(urllib.parse.urlparse(callback_url))
            parsed[1] = gcs_manager_domain
            callback_url = urllib.parse.urlunparse(parsed)

        return callback_url

    @staticmethod
    def negotiate_error_response(request, error):
        if "text/html" == request.accept_mimetypes.best:
            response = make_response(
                render_template(
                    "error.jinja2", status_code=error.status_code, **error.to_dict()
                )
            )
        else:
            error_dict = error.to_dict()
            error_dict.pop("detail")
            response = jsonify(error_dict)

        response.status_code = error.status_code

        return response

    @classmethod
    def handle_validation_error(cls, error):
        return cls.negotiate_error_response(request, error)
