import json
from typing import List

import oauth2client
from oauth2client.client import HttpAccessTokenRefreshError
from oauth2client import service_account
from oauth2client import GOOGLE_TOKEN_URI
from flask import current_app
import httplib2

import globus.manager.api.errors as errors
from globus.manager.utils.update_query_string import update_query_string
from .oidc_user_credential import OIDCUserCredential
from globus.manager.constants import ConfigurationObjectTag


class GoogleCloudStorageUserCredential(OIDCUserCredential):
    POLICY_DATA_TYPE = "google_cloud_storage_user_credential_policies#1.0.0"
    POLICY_DATA_TYPES = [POLICY_DATA_TYPE]
    CONNECTOR_ID = "56366b96-ac98-11e9-abac-9cb6d0d9fd63"
    SERVICE_SCOPES = (
        "email "
        "profile "
        "https://www.googleapis.com/auth/devstorage.read_write "
        "https://www.googleapis.com/auth/cloudplatformprojects.readonly "
    )

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.USER_CREDENTIAL.name + CONNECTOR_ID
        )
    }

    @property
    def connector_properties(self) -> List[str]:
        return super().connector_properties + ["projects"]

    def _get_google_cred(self):
        cred = None
        if self.uses_service_auth:
            scopes = self.SERVICE_SCOPES
            keyfile_dict = self.storage_gateway.service_account_key
            cred = service_account.ServiceAccountCredentials.from_json_keyfile_dict(
                keyfile_dict,
                scopes,
            )

        else:
            refresh_token = self.refresh_token
            secret = self.storage_gateway.secret
            client_id = self.storage_gateway.client_id
            cred = oauth2client.client.GoogleCredentials(
                None,
                client_id,
                secret,
                refresh_token,
                None,
                "https://accounts.google.com/o/oauth2/token",
                "globus-connect-server-manager",
            )
        return cred

    def _get_service_access_token(self):
        try:
            gcred = self._get_google_cred()
            gcred.refresh(httplib2.Http())
        except Exception as e:
            current_app.logger.warning("Service Credential grant failure: %s", str(e))
            raise errors.InvalidCredentialError(user_credential_id=self.id)

        return gcred.access_token, gcred.token_expiry

    def __init__(
        self, appsync_data=None, doc=None, storage_gateway=None, credentials=None
    ):
        if doc:
            self.username = doc["username"]
            new_doc = {**doc}
            new_doc.pop("display_name", None)
            # Policies must be either empty (for an unprovisioned credential)
            # or passed in as a credentials object from the google callback
            policies = new_doc.get("policies", None)
            if policies is not None and policies != {
                "DATA_TYPE": self.POLICY_DATA_TYPE
            }:
                raise ValueError("policies")
            doc = new_doc

        super().__init__(
            appsync_data=appsync_data, doc=doc, storage_gateway=storage_gateway
        )
        if doc and self.uses_service_auth:
            service_dict = self.storage_gateway.service_account_key
            access_token, expiry = self._get_service_access_token()
            p = doc.get("policies", {})
            policies = {
                **p,
                **{
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                    "access_token": access_token,
                    "token_expiry": expiry,
                    "refresh_token": "service",
                    "sub": service_dict["client_id"],
                    "email": self.username,
                    "scopes": self.SERVICE_SCOPES,
                },
            }
            doc = {
                **doc,
                **{
                    "display_name": service_dict["client_id"],
                    "invalid": False,
                    "deleted": False,
                    "provisioned": True,
                    "policies": policies,
                },
            }
            self._patch(**doc)
        elif doc:
            patch = self._credentials_to_patch(credentials)
            patch["deleted"] = False
            patch["provisioned"] = bool(credentials)
            self._patch(**patch)

    @property
    def connector_id(self):
        return self.CONNECTOR_ID

    @connector_id.setter
    def connector_id(self, value):
        pass

    def can_access_project(self, project):
        """
        Confirm that the project is viewable by this user credential
        """
        # confirm project is accessible by user_credential
        cred = self._get_google_cred()
        http = cred.authorize(httplib2.Http())
        url = "https://cloudresourcemanager.googleapis.com/v1/projects/{}"
        try:
            response, content = http.request(url.format(project))
        except oauth2client.client.HttpAccessTokenRefreshError as e:
            current_app.logger.exception("Error checking project access permissions", e)
            if not self.uses_service_auth:
                self.invalid = True
            raise errors.InvalidCredentialError(
                "Globus Connnect Server cannot verify that your credential is "
                "able to access Google Cloud Storage. Please contact the "
                "endpoint administrator to verify the storage gateway "
                "project configuration.",
                user_credential_id=self.id,
            )

        if int(response.get("status", 0)) != 200:
            try:
                data = json.loads(content.decode("utf8"))
                current_app.logger.error(
                    "Error listing Google projects: %s: %s",
                    data["code"],
                    data["error"]["message"],
                )
            except Exception:
                if not self.uses_service_auth:
                    self.invalid = True
                current_app.logger.error(
                    "Error listing Google projects: %s", response.get("status", 0)
                )
                raise errors.InvalidCredentialError(
                    "Globus Connnect Server cannot verify that your credential is "
                    "able to access Google Cloud Storage. Please contact the "
                    "endpoint administrator to verify the storage gateway "
                    "project configuration.",
                    user_credential_id=self.id,
                )

            raise errors.ForbiddenError(
                "Your credentials do not have permissions to access the Google project {}".format(
                    project,
                )
            )
        else:
            # sanity check we got back the expected project
            data = json.loads(content.decode("utf8"))
            if data.get("projectId") != project:
                raise errors.InternalServerError("Unexpected Google Cloud response")
        return True

    @property
    def projects(self):
        if not hasattr(self, "_projects") or self._projects is None:
            if self.invalid:
                raise AttributeError(
                    "The projects property is not defined for this credential"
                )

            current_app.logger.debug("Looking up projects")
            try:
                cred = self._get_google_cred()
                http = cred.authorize(httplib2.Http())
            except Exception as e:
                current_app.logger.warning("invalid credentials: %s", str(e))
                if not self.uses_service_auth:
                    self.invalid = True
                raise AttributeError(
                    "The projects property is not defined for this credential"
                )

            final_page = False
            page_token = None
            projects = []
            while not final_page:
                url = "https://cloudresourcemanager.googleapis.com/v1/projects"
                if page_token:
                    url = update_query_string(
                        url,
                        dict(
                            page_token=page_token,
                            fields="nextPageToken,projects(projectId,name))",
                        ),
                    )
                try:
                    response, content = http.request(url)
                    status = int(response.get("status", 0))
                except HttpAccessTokenRefreshError as e:
                    current_app.logger.warning("invalid credentials: %s", str(e))
                    if not self.uses_service_auth:
                        self.invalid = True
                    raise AttributeError(
                        "The projects property is not defined for this credential"
                    )
                except Exception as e:
                    current_app.logger.warning("project enumeration error: %s", str(e))
                    raise AttributeError(
                        "The projects property is not defined for this credential"
                    )

                if status == 200:
                    data = json.loads(content.decode("utf8"))
                    page_token = data.get("nextPageToken")
                    final_page = page_token is None
                    projects.extend(data.get("projects", []))
                # if we get back a 4xx error, the credential is no longer valid
                elif status == 401:
                    if not self.uses_service_auth:
                        self.invalid = True
                    projects = []
                    break
                # other errors could be temporary so we don't invalidate the
                # cred
                else:
                    projects = []
                    break

            # If google_cloud_storage_projects in the Storage Gateway config,
            # filter out any projects that are not in that list
            try:
                allowed_projects = self.storage_gateway.projects
            except AttributeError:
                allowed_projects = []
            if allowed_projects:
                projects = list(
                    [p for p in projects if p["projectId"] in allowed_projects]
                )
            self._projects = projects

        return [
            {"projectId": p["projectId"], "name": p["name"]} for p in self._projects
        ]

    @projects.deleter
    def projects(self):
        self._projects = None

    @projects.setter
    def projects(self, value):
        pass

    @property
    def uses_service_auth(self) -> bool:
        return not self.storage_gateway.user_credential_required

    @property
    def supports_revoke(self) -> bool:
        return not self.uses_service_auth

    def refresh(self):
        if self.uses_service_auth:
            service_dict = self.storage_gateway.service_account_key
            access_token, expiry = self._get_service_access_token()
            self.refresh_token = "service"
            self.access_token = access_token
            self.token_expiry = expiry
            self.sub = service_dict["client_id"]
            self.display_name = service_dict["client_id"]
            self.email = self.username
            self.scopes = self.SERVICE_SCOPES
            self.invalid = False
            self.provisioned = True
            self.configuration_hash.changed()
            current_app.logger.debug("Token refreshed: {}".format(self.token_expiry))
        else:
            super().refresh()

    @property
    def token_uri(self) -> str:
        return GOOGLE_TOKEN_URI

    def __repr__(self):
        return "<GoogleCloudStorageUserCredential {}>".format(self.id)
