from collections.abc import Sequence
from typing import List, Optional

from .collection import Collection
from .decorators import checks_property
from globus.manager.constants import ConfigurationObjectTag


class PosixCollection(Collection):
    POLICY_DATA_TYPE = "posix_collection_policies#1.1.0"
    CONNECTOR_POLICIES = {
        "posix_collection_policies#1.0.0": [],
        "posix_collection_policies#1.1.0": [
            "sharing_groups_allow",
            "sharing_groups_deny",
        ],
    }
    POLICY_DATA_TYPES = list(CONNECTOR_POLICIES.keys())

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.COLLECTION.name
            + "145812c8-decc-41f1-83cf-bb2a85a2a70b"
        )
    }

    @property
    def connector_id(self) -> str:
        return "145812c8-decc-41f1-83cf-bb2a85a2a70b"

    @connector_id.setter
    def connector_id(self, value):
        pass

    @property
    def sharing_groups_allow(self) -> List[str]:
        """
        List of connector-specific usernames allowed to create new
        guest collections on this mapped collection. (**Added in
        API 1.2.0**)
        """
        return self.get_policy("sharing_groups_allow")

    @sharing_groups_allow.setter
    def sharing_groups_allow(self, value: Optional[List[str]]):
        self.set_policy("sharing_groups_allow", value)

    @sharing_groups_allow.deleter
    def sharing_groups_allow(self):
        self.delete_policy("sharing_groups_allow")

    @checks_property(prop="sharing_groups_allow")
    def _check_sharing_groups_allow_value(self, value: Optional[List[str]]):
        if self.is_guest_collection:
            raise AttributeError()
        self.__validate_string_list("sharing_groups_allow", value)

    @property
    def sharing_groups_deny(self) -> Optional[List[str]]:
        """
        List of connector-specific usernames denyed to create new
        guest collections on this mapped collection. (**Added in
        API 1.2.0**)
        """
        return self.get_policy("sharing_groups_deny")

    @sharing_groups_deny.setter
    def sharing_groups_deny(self, value: Optional[List[str]]):
        self.set_policy("sharing_groups_deny", value)

    @sharing_groups_deny.deleter
    def sharing_groups_deny(self):
        self.delete_policy("sharing_groups_deny")

    @checks_property(prop="sharing_users_deny")
    def _check_sharing_users_deny_value(self, value: Optional[List[str]]):
        if self.is_guest_collection:
            raise AttributeError()
        self.__validate_string_list("sharing_users_deny", value)

    def check_sharing_user_allowed(self, username: str) -> bool:
        """
        Check whether a user is allowed to create a guest collection, based on a
        mapped collection's policies.

        This uses the same algorithm as the guest collection's check_user_allowed,
        but populates the lists with the contents of the sharing_users_allow,
        sharing_users_deny, sharing_groups_allow, and sharing_groups_deny
        properties.

        :param str username:
            Connector-specific username to check for access.
        :returns:
            True if the user can create a guest collection, False otherwise.
        """
        if self.is_guest_collection or not self.allow_guest_collections:
            return False

        if not self.storage_gateway.check_user_allowed(username):
            return False

        return self.storage_gateway._check_user_allowed(
            username,
            getattr(self, "sharing_users_allow", None),
            getattr(self, "sharing_users_deny", None),
            getattr(self, "sharing_groups_allow", None),
            getattr(self, "sharing_groups_deny", None),
        )

    @staticmethod
    def __validate_string_list(name, value: Optional[List[str]]):
        if value:
            if not isinstance(value, Sequence) or isinstance(value, str):
                raise ValueError("Invalid value for the {} property.".format(name))
            for v in value:
                if not isinstance(v, str):
                    raise ValueError("Invalid value for the {} property.".format(name))

    def __repr__(self):
        return "<PosixCollection {}>".format(self.id)
