from urllib.parse import urlparse
from collections.abc import Sequence
from typing import List

from .connector import Connector
from .decorators import checks_policy_property
from .storage_gateway import StorageGateway
from globus.manager.constants import ConfigurationObjectTag


class S3StorageGateway(StorageGateway):
    CONNECTOR_ID = "7643e831-5f6c-4b47-a07f-8ee90f401d23"
    POLICY_DATA_TYPE = "s3_storage_policies#1.1.0"
    CONNECTOR_POLICIES = {
        "s3_storage_policies#1.0.0": [
            "s3_endpoint",
            "s3_buckets",
            "s3_user_credential_required",
        ],
        "s3_storage_policies#1.1.0": [
            "s3_endpoint",
            "s3_buckets",
            "s3_user_credential_required",
            "s3_requester_pays",
        ],
    }
    POLICY_DATA_TYPES = list(CONNECTOR_POLICIES.keys())
    connector_policy_names = CONNECTOR_POLICIES[POLICY_DATA_TYPE]

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }

    connector = Connector(id=CONNECTOR_ID, display_name="S3")

    s3_endpoint = StorageGateway.connector_property(
        "s3_endpoint", required=True, doc=" URL of the S3 API endpoint"
    )

    @checks_policy_property(prop="s3_endpoint", required=True)
    def check_s3_endpoint(self, value: str):
        if value is not None:
            try:
                parsed_value = urlparse(value)
                if parsed_value.scheme not in ["http", "https"]:
                    raise ValueError()
            except Exception:
                raise ValueError()
        else:
            raise ValueError()

    s3_buckets = StorageGateway.connector_property(
        "s3_buckets",
        doc="Optional list of S3 buckets which this storage gateway allows "
        "access to. If this is not set, then this Storage Gateway allows access "
        "to any buckets that User Credentials have access to.",
    )

    @checks_policy_property(prop="s3_buckets")
    def check_s3_buckets(self, value: List[str]):
        if value is not None:
            if not isinstance(value, Sequence) or isinstance(value, str):
                raise ValueError()
            for v in value:
                if not isinstance(v, str):
                    raise ValueError()

    s3_user_credential_required = StorageGateway.connector_property(
        "s3_user_credential_required", required=True
    )

    @checks_policy_property(prop="s3_user_credential_required", required=True)
    def check_s3_user_credential_required(self, value: bool):
        if value is None or not isinstance(value, bool):
            raise ValueError()

    s3_requester_pays = StorageGateway.connector_property("s3_requester_pays")

    @checks_policy_property(prop="s3_requester_pays")
    def check_s3_requester_pays(self, value: bool):
        if value is not None and not isinstance(value, bool):
            raise ValueError()

    @property
    def load_dsi_module(self):
        return "s3"

    @load_dsi_module.setter
    def load_dsi_module(self, value):
        pass

    @property
    def process_user(self):
        return "gcsweb"

    @process_user.setter
    def process_user(self, value):
        pass

    @property
    def require_credential(self):
        return self.s3_user_credential_required

    def default_mapping(self, identity_data):
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return identity_data["username"]

    def __repr__(self):
        return "<S3StorageGateway {}>".format(self.id)
