from typing import List

import marshmallow as ma

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeField, DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


BLACKPEARL_CONNECTOR_ID = "7e3f3f5e-350c-4717-891a-2f451c24b0d4"


class BlackPearlStoragePolicies(ma.Schema):
    """
    Connector-specific storage gateway policies for the Blackpearl connector
    """

    _DATA_TYPE = "blackpearl_storage_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    s3_endpoint = ma.fields.Url(
        schemes=["http", "https"],
        description="""
        The URL of the S3 endpoint of the BlackPearl appliance
        to use to access collections on this Storage Gateway.
        """,
    )
    bp_access_id_file = ma.fields.Str(
        description="""
        Path to the file which provides mappings from usernames within the
        configured identity domain to the ID and secret associated with the
        user's BlackPearl account
        """
    )
    private = ["bp_access_id_file"]
    connector_id = BLACKPEARL_CONNECTOR_ID


class BlackPearlCollectionPolicies(ma.Schema):
    """
    Connector-specific collection policies for the BlackPearl connector
    """

    _DATA_TYPE = "blackpearl_collection_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    private = []  # type: List[str]
    connector_id = BLACKPEARL_CONNECTOR_ID


class BlackPearlUserCredentialPolicies(ma.Schema):
    """
    Connector-specific user credential policies for the Blackpearl connector
    """

    _DATA_TYPE = "blackpearl_user_credential_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    access_id = ma.fields.String(dump_only=True, description="BlackPearl access id")
    secret_key = ma.fields.String(dump_only=True, description="BlackPearl secret key")
    private = ["secret_key"]  # type: List[str]
    connector_id = BLACKPEARL_CONNECTOR_ID


blackpearl_user_credential_policies_private_schema = BlackPearlUserCredentialPolicies()
blackpearl_user_credential_policies_schema = BlackPearlUserCredentialPolicies(
    load_only=BlackPearlUserCredentialPolicies.private
)
blackpearl_storage_policies_include_private_schema = BlackPearlStoragePolicies()
blackpearl_storage_policies_schema = BlackPearlStoragePolicies(
    load_only=BlackPearlStoragePolicies.private
)
blackpearl_storage_policies_private_schema = BlackPearlStoragePolicies()
blackpearl_collection_policies_schema = BlackPearlCollectionPolicies()
blackpearl_collection_policies_private_schema = BlackPearlCollectionPolicies(
    load_only=BlackPearlCollectionPolicies.private
)

DataTypeSchema.register_schema(blackpearl_storage_policies_schema)
DataTypeSchema.register_schema(blackpearl_user_credential_policies_schema)
DataTypeSchema.register_schema(blackpearl_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=blackpearl_storage_policies_schema,
    private=blackpearl_storage_policies_include_private_schema,
)

UserCredentialSchema.register_schema(
    public=blackpearl_user_credential_policies_schema,
    private=blackpearl_user_credential_policies_private_schema,
)

CollectionSchema.register_schema(
    public=blackpearl_collection_policies_schema,
    private=blackpearl_collection_policies_schema,
)
