from typing import List

import marshmallow as ma

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeField, DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


GOOGLE_DRIVE_CONNECTOR_ID = "976cf0cf-78c3-4aab-82d2-7c16adbcc281"


class GoogleDriveStoragePolicies(ma.Schema):
    """
    Connector-specific storage gateway policies for the Google Drive connector
    """

    _DATA_TYPE = "google_drive_storage_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    client_id = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        Client ID registered with the Google Application console to access
        Google Drive.
        """,
    )
    secret = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        Secret created to access access Google Drive with the client_id in this
        policy.
        """,
    )
    user_api_rate_quota = ma.fields.Int(
        required=False, description="User API Rate quota associated with this client ID"
    )
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        description="""
        URL of the auth callback that must be registered on the Google API
        console for the application client_id in order to process Google
        credentials.
        """,
    )

    private = ["client_id", "secret", "user_api_rate_quota"]
    connector_id = GOOGLE_DRIVE_CONNECTOR_ID

    @ma.pre_load
    def ignore_dump_only(self, data, **kwargs):
        """
        Ignore dump_only schema fields when loading.
        """
        for k, v in self.fields.items():
            if v.dump_only:
                data.pop(k, None)
        return data


class GoogleDriveCollectionPolicies(ma.Schema):
    """
    Connector-specific collection policies for the Google Drive connector
    """

    _DATA_TYPE = "google_drive_collection_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    private = []  # type: List[str]
    connector_id = GOOGLE_DRIVE_CONNECTOR_ID


class GoogleDriveUserCredentialPolicies(ma.Schema):
    """
    Connector-specific user credential policies for the Google Drive connector
    """

    _DATA_TYPE = "google_drive_user_credential_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    sub = ma.fields.Str(allow_none=True, description="OAuth subject identifier claim")
    email = ma.fields.Email(allow_none=True, description="OAuth email claim")
    access_token = ma.fields.Str(
        required=False,
        allow_none=True,
        description="OAuth access token",
    )
    refresh_token = ma.fields.Str(
        required=False, allow_none=True, description="OAuth refresh token"
    )
    token_expiry = ma.fields.AwareDateTime(
        description="OAuth access token expiration time"
    )
    scopes = ma.fields.List(
        ma.fields.Str,
        description="OAuth scopes associated with this access token",
        example=[
            "email",
            "profile",
            "https://www.googleapis.com/auth/drive",
            "https://www.googleapis.com/auth/drive.appfolder",
        ],
    )

    private = ["access_token", "refresh_token"]
    connector_id = GOOGLE_DRIVE_CONNECTOR_ID


google_drive_user_credential_policies_private_schema = (
    GoogleDriveUserCredentialPolicies()
)
google_drive_user_credential_policies_schema = GoogleDriveUserCredentialPolicies(
    load_only=GoogleDriveUserCredentialPolicies.private
)
google_drive_storage_policies_private_schema = GoogleDriveStoragePolicies()
google_drive_storage_policies_schema = GoogleDriveStoragePolicies(
    load_only=GoogleDriveStoragePolicies.private
)
google_drive_collection_policies_schema = GoogleDriveCollectionPolicies(
    load_only=GoogleDriveCollectionPolicies.private
)
google_drive_collection_policies_private_schema = GoogleDriveCollectionPolicies()

DataTypeSchema.register_schema(google_drive_storage_policies_schema)
DataTypeSchema.register_schema(google_drive_user_credential_policies_schema)
DataTypeSchema.register_schema(google_drive_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=google_drive_storage_policies_schema,
    private=google_drive_storage_policies_private_schema,
)

UserCredentialSchema.register_schema(
    public=google_drive_user_credential_policies_schema,
    private=google_drive_user_credential_policies_private_schema,
)

CollectionSchema.register_schema(
    public=google_drive_collection_policies_schema,
    private=google_drive_collection_policies_private_schema,
)
