from typing import List

import marshmallow as ma
from marshmallow_oneofschema import OneOfSchema

from ..cleanup_none_mixin import CleanupNoneMixin
from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeField, DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


POSIX_CONNECTOR_ID = "145812c8-decc-41f1-83cf-bb2a85a2a70b"


class PosixStoragePolicies(ma.Schema, CleanupNoneMixin):
    """
    Connector-specific storage gateway policies for the POSIX connector.
    """

    _DATA_TYPE = "posix_storage_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    groups_allow = ma.fields.List(
        ma.fields.Str,
        allow_none=True,
        description="""
        List of POSIX group names allowed to access this storage gateway
        """,
    )
    groups_deny = ma.fields.List(
        ma.fields.Str,
        allow_none=True,
        description="""
        List of POSIX group names denied access this storage gateway
        """,
    )

    private = ["groups_allow", "groups_deny"]
    connector_id = POSIX_CONNECTOR_ID


class PosixCollectionPolicies_1_0_0(ma.Schema):
    """
    Connector-specific collection policies for the POSIX connector
    """

    _DATA_TYPE = "posix_collection_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    private = []  # type: List[str]
    connector_id = POSIX_CONNECTOR_ID


class PosixCollectionPolicies_1_1_0(PosixCollectionPolicies_1_0_0):
    """
    Connector-specific collection policies for the POSIX connector
    """

    _DATA_TYPE = "posix_collection_policies#1.1.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    sharing_groups_allow = ma.fields.List(
        ma.fields.Str,
        allow_none=True,
        description="""
        List of POSIX group names allowed to create shares on this collection
        """,
    )
    sharing_groups_deny = ma.fields.List(
        ma.fields.Str,
        allow_none=True,
        description="""
        List of POSIX group names denied access to create shares on this
        collection.
        """,
    )
    private = ["sharing_groups_allow", "sharing_groups_deny"]
    connector_id = POSIX_CONNECTOR_ID


class PosixUserCredentialPolicies(ma.Schema):
    """
    Connector-specific user credential policies for the POSIX connector
    """

    _DATA_TYPE = "posix_user_credential_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    connector_id = POSIX_CONNECTOR_ID
    private = []  # type: List[str]


posix_storage_policies_schema = PosixStoragePolicies(
    load_only=PosixStoragePolicies.private
)
posix_storage_policies_private_schema = PosixStoragePolicies()
posix_user_credential_policies_schema = PosixUserCredentialPolicies()

posix_collection_policies_1_0_0_schema = PosixCollectionPolicies_1_0_0(
    load_only=PosixCollectionPolicies_1_0_0.private
)
posix_collection_policies_1_1_0_schema = PosixCollectionPolicies_1_1_0(
    load_only=PosixCollectionPolicies_1_1_0.private
)

posix_collection_policies_1_0_0_private_schema = PosixCollectionPolicies_1_0_0()
posix_collection_policies_1_1_0_private_schema = PosixCollectionPolicies_1_1_0()


class PosixCollectionPolicies(OneOfSchema):
    """
    Connector-specific collection policies for the POSIX connector
    """

    _DATA_TYPE = posix_collection_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "posix_collection_policies#1.0.0": posix_collection_policies_1_0_0_schema,
        "posix_collection_policies#1.1.0": posix_collection_policies_1_1_0_schema,
    }
    type_field_remove = False
    connector_id = POSIX_CONNECTOR_ID
    title = "PosixCollectionPolicies"

    def get_obj_type(self, obj):
        return self._DATA_TYPE

    private = posix_collection_policies_1_1_0_schema.private + []


class PosixCollectionPoliciesPrivate(OneOfSchema):
    _DATA_TYPE = posix_collection_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "posix_collection_policies#1.0.0": posix_collection_policies_1_0_0_private_schema,
        "posix_collection_policies#1.1.0": posix_collection_policies_1_1_0_private_schema,
    }
    title = "PosixCollectionPolicies"
    type_field_remove = False
    connector_id = POSIX_CONNECTOR_ID

    def get_obj_type(self, obj):
        return self._DATA_TYPE


posix_collection_policies_schema = PosixCollectionPolicies(
    load_only=PosixCollectionPolicies.private
)
posix_collection_policies_private_schema = PosixCollectionPoliciesPrivate()

DataTypeSchema.register_schema(posix_storage_policies_schema)
DataTypeSchema.register_schema(posix_user_credential_policies_schema)
DataTypeSchema.register_schema(posix_collection_policies_1_0_0_schema)
DataTypeSchema.register_schema(posix_collection_policies_1_1_0_schema)

StorageGatewaySchema.register_schema(
    public=posix_storage_policies_schema, private=posix_storage_policies_private_schema
)

UserCredentialSchema.register_schema(public=posix_user_credential_policies_schema)

CollectionSchema.register_schema(
    public=posix_collection_policies_schema,
    private=posix_collection_policies_private_schema,
)
