from typing import List

import marshmallow as ma

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeField, DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema
from .posix import (
    PosixStoragePolicies,
    PosixUserCredentialPolicies,
    PosixCollectionPolicies_1_1_0,
)


POSIX_STAGING_CONNECTOR_ID = "052be037-7dda-4d20-b163-3077314dc3e6"


class PosixStagingEnvironment(ma.Schema):
    """
    Variables to set in the environment when executing
    the stage_app.
    """

    name = ma.fields.String(required=True, description="Environment variable name")
    value = ma.fields.String(required=True, description="Environment variable value")


class PosixStagingStoragePolicies(PosixStoragePolicies):
    """
    Connector-specific storage gateway policies for the POSIX Staging connector
    """

    _DATA_TYPE = "posix_staging_storage_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    stage_app = ma.fields.String(required=True, description="Path to the stage app")
    environment = ma.fields.List(
        ma.fields.Nested(PosixStagingEnvironment),
        description="Variables to set in the environment when executing the stage_app",
    )

    private = PosixStoragePolicies.private + ["stage_app", "environment"]
    connector_id = POSIX_STAGING_CONNECTOR_ID


class PosixStagingCollectionPolicies(PosixCollectionPolicies_1_1_0):
    """
    Connector-specific collection policies for the POSIX Staging connector
    """

    _DATA_TYPE = "posix_staging_collection_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    private = PosixCollectionPolicies_1_1_0.private + []  # type: List[str]
    connector_id = POSIX_STAGING_CONNECTOR_ID


class PosixStagingUserCredentialPolicies(PosixUserCredentialPolicies):
    """
    Connector-specific user credential policies for the POSIX Staging connector
    """

    _DATA_TYPE = "posix_staging_user_credential_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    connector_id = POSIX_STAGING_CONNECTOR_ID
    private = PosixUserCredentialPolicies.private + []  # type: List[str]


posix_staging_storage_policies_schema = PosixStagingStoragePolicies(
    load_only=PosixStagingStoragePolicies.private
)
posix_staging_storage_policies_private_schema = PosixStagingStoragePolicies()
posix_staging_user_credential_policies_schema = PosixStagingUserCredentialPolicies()
posix_staging_collection_policies_schema = PosixStagingCollectionPolicies(
    load_only=PosixStagingCollectionPolicies.private,
)
posix_staging_collection_policies_private_schema = PosixStagingCollectionPolicies()

DataTypeSchema.register_schema(posix_staging_storage_policies_schema)
DataTypeSchema.register_schema(posix_staging_user_credential_policies_schema)
DataTypeSchema.register_schema(posix_staging_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=posix_staging_storage_policies_schema,
    private=posix_staging_storage_policies_private_schema,
)

UserCredentialSchema.register_schema(
    public=posix_staging_user_credential_policies_schema
)

CollectionSchema.register_schema(
    public=posix_staging_collection_policies_schema,
    private=posix_staging_collection_policies_private_schema,
)
