import threading
from contextlib import contextmanager

from globus.manager.utils.interprocess_lock import InterprocessLock


class ExclusiveSubscriptionLock:
    """
    Provides a subscription lock that is both inter and intra process safe.
    """

    def __init__(self, subscription_lock: InterprocessLock) -> None:
        """
        Initiaze the intra+inter process subscription lock.

        The subscription_lock supplied by appsync is interprocess safe; it will
        provide the locking necessary when accessing the database with a single
        thread within a single process. The assistant requires intraprocess locking
        as well since multiple threads need to synchronize.
        """
        self.__subscription_lock = subscription_lock
        # RLock is reentrant allowing the same thread to acquire the lock even if
        # the thread already acquired it in a calling function.
        self.__intra_process_lock = threading.RLock()

    @contextmanager
    def exclusive(self):
        """
        Grab the intra+inter process subscription lock.
        """
        with self.__intra_process_lock, self.__subscription_lock.exclusive():
            yield
