from contextlib import suppress
import datetime
import os
import sched
import shutil
import threading
from typing import Any, Dict, List, Optional

import flask
import marshmallow as ma

import globus.manager.api.models as models
from globus.manager.api.schemas.appsync.appsync_data import appsync_data_schema
from globus.manager.constants import ConfigurationObjectTag
from globus_gcs.gridftp.gridftp_service import GridFTPService
from .exclusive_subscription_lock import ExclusiveSubscriptionLock


class NotificationManager:
    """
    This class handles maintaining synchronization between appsync and the
    local sqlite database.
    """

    PAGINATION_DELAY = 5
    """
    Delay (in seconds) to wait after a query returns a paginated response to
    perform another query.
    """

    ERROR_DELAY = 30
    """
    Delay (in seconds) to wait after a query or subscription call fails before
    trying again.
    """

    RELOAD_PATH = os.path.expanduser("~gcsweb/etc/httpd/reload/vhost_change")

    vhost_dir = os.path.expanduser("~gcsweb/etc/httpd/conf.d/")

    def __init__(
        self,
        scheduler: sched.scheduler,
        app: flask.Flask,
        subscription_lock: ExclusiveSubscriptionLock,
    ):
        """
        Initialize a NotificationManager object to handle an appsync subscription.

        :param sched.scheduler scheduler:
            A scheduler which this class will use to register timed events.
        :param GCSManagerApp app:
            An app which this class will use to interact with appsync, the
            database, locks and logs.
        :param: subscription_lock
            Lock to use when processing query results. This locks out the GCS
            manager process from processing requests that would interact with
            the database while we're processing changes that might make many
            changes to the database.
        """
        self.scheduler = scheduler
        "Scheduler to register events with"

        self.app = app
        "Flask application this object is a part of"

        self.__subscription_lock = subscription_lock
        "Intra+Inter process subscription (db) lock"

        self.deferred_items = []  # type: List[Dict[str, Any]]
        """
        List of appsync items which can not yet be processed because they
        have a foreign key reference to some other object which we don't
        (yet) have in the database. Ordering of notifications isn't well
        defined in appsync
        """

        self.last_item_timestamp = None  # type: Optional[datetime.datetime]
        """
        Highest value of the updated_at property for an object we've
        received in response to an appsync query or notification. This is
        used to make resubscriptions only query for recent changes.
        """
        self.appsync_client = app.appsync_client

        self.page_token_timestamp = None  # type: Optional[datetime.datetime]
        """
        Value of self.last_item_timestamp when a pagination query
        began. This value cannot change while pagination is in progress.
        """

        self.page_token = None  # type: Optional[str]
        """
        Query response pagination token.
        """

        self.vhost_dirty = False  # type: bool
        """
        Flag indicating whether we've changed any vhosts since the last time
        we've touched the vhost reload directory.
        """

        self.vhost_timestamp = datetime.datetime.now()
        """
        Timestamp of the last change to any vhost managed by this process.
        """

        self.vhost_reload = None  # type: Optional[sched.Event]
        """
        ID of a scheduler event to touch the reload.
        """

        self.notifier = None
        """
        Value returned from the appsync client's subscribe method.
        """

        self.mutex = threading.Lock()
        """
        Mutex to guard vhost_dirty, vhost_timestamp, and notifier
        """

        self.scheduler.enter(0, 0, self.query_and_subscribe)

    def query_and_subscribe(self) -> None:
        """
        Event to resynchronize state in the local database with the appsync
        database. It performs a full query of the objects in appsync, and
        then subscribes for changes to the appsync endpoint.

        This is triggered when the NotificationManager is first started, and
        then also when the appsync subscription is disconnected (which happens
        approximately every 24 hours).

        To be easy on our appsync implementation, when a paginated result
        is returned from the appsync query, the next page request is slightly
        delayed to a further event.

        If there's an error querying or subscribing, a longer delay is used.
        If this function returns successfully, then a new event is registered,
        either another call to this event, or via the expected eventual
        on_disconnect call.
        """
        with self.app.app_context(), self.__subscription_lock.exclusive():
            self.app.logger.debug("query_and_subscribe enter")
            old_subscription: Optional[str] = None
            try:
                endpoint: Optional[
                    models.Endpoint
                ] = models.Endpoint.query.one_or_none()
                if endpoint:
                    old_subscription = endpoint.subscription_id
            except Exception as e:
                self.app.logger.warning(
                    "Error checking endpoint subscription_id: %s", str(e)
                )
                pass

            has_next_page = True

            since = None
            if self.page_token:
                since = self.page_token_timestamp
            elif self.last_item_timestamp:
                since = self.last_item_timestamp - datetime.timedelta(minutes=5)
                self.page_token_timestamp = since
            while has_next_page:
                try:
                    items, self.page_token = self.appsync_client.query(
                        since=since, page_token=self.page_token
                    )
                    self.process_items(items, old_subscription)
                    has_next_page = bool(self.page_token)
                except Exception as e:
                    self.app.logger.warning("Error syncing state: %s", str(e))
                    self.scheduler.enter(self.ERROR_DELAY, 0, self.query_and_subscribe)
                    return
                if self.page_token:
                    # If we're hitting pagination issues, let's take it easy
                    # on appsync
                    self.scheduler.enter(
                        self.PAGINATION_DELAY, 0, self.query_and_subscribe
                    )
                    return

            try:
                with self.mutex:
                    if not self.notifier:
                        self.notifier = self.appsync_client.subscribe(
                            on_message=self.on_message, on_disconnect=self.on_disconnect
                        )
                    if self.vhost_dirty and not self.vhost_reload:
                        self.vhost_reload = self.scheduler.enter(
                            1, 0, self.touch_reload
                        )
            except Exception as e:
                self.app.logger.error("Error processing subscription %s", str(e))

            with self.mutex:
                if not self.notifier:
                    self.app.logger.warning("Unable to subscribe for notifications")
                    self.scheduler.enter(self.ERROR_DELAY, 0, self.query_and_subscribe)

    def process_items(
        self,
        items: List[Dict[str, Any]],
        old_subscription: Optional[str],
        ignore_old_updates: bool = False,
    ) -> None:
        """
        Add items from the appsync response or change notification to the local
        database, updating local configuration files if necessary.

        If there are any deferred changes from a previous query or notifications,
        they may be processed if their foreign keys are now valid. Any objects
        which can not be processed with the current database state will be
        stashed in self.deferred_items for later processing.

        :param items: List of items returned from appsync
        :param old_subscription:
            Subscription ID of the endpoint before the query result or
            notification data has been processed. This is used to update
            configurations that depend on whether the endpoint has a
            subscription or not.
        """
        items.extend(self.deferred_items)
        changed_collection_ids = [
            i["id"] for i in items if i["tag"] == ConfigurationObjectTag.COLLECTION
        ]
        changed_tags = set(i["tag"] for i in items)

        changed_endpoint = bool(ConfigurationObjectTag.ENDPOINT in changed_tags)
        changed_certificate = bool(ConfigurationObjectTag.CERTIFICATE in changed_tags)

        if items:
            newest_change = sorted([i["updated_at"] for i in items], reverse=True)[0]
            if (
                self.last_item_timestamp is None
                or newest_change > self.last_item_timestamp
            ):
                self.last_item_timestamp = newest_change

        # Update models in local database. Note that anything we hard delete will
        # not show up in queries below.
        with self.__subscription_lock.exclusive():
            self.deferred_items = self.appsync_client.commit_items_to_session(
                models.db.session, items, ignore_old_updates
            )

        endpoint: Optional[models.Endpoint] = None
        try:
            endpoint = models.Endpoint.query.one_or_none()
        except Exception as e:
            self.app.logger.warning("Unable to update the endpoint: %s", str(e))
        changed_port = False

        if endpoint is not None:
            saved_control_port_config: Optional[
                models.Configuration
            ] = models.Configuration.query.filter_by(
                name="gridftp_control_channel_port"
            ).one_or_none()
            if saved_control_port_config is None:
                saved_control_port_config = models.Configuration(
                    "gridftp_control_channel_port",
                    str(endpoint.gridftp_control_channel_port),
                )
                changed_port = True
            elif endpoint.gridftp_control_channel_port != int(
                saved_control_port_config.value
            ):
                saved_control_port_config.value = str(
                    endpoint.gridftp_control_channel_port
                )
                changed_port = True
            if changed_port:
                models.db.session.add(saved_control_port_config)
                models.db.session.commit()

        if bool(ConfigurationObjectTag.WILDCARD_DOMAIN in changed_tags) or changed_port:
            node_id: str = (
                models.Configuration.query.filter_by(name="node_id").one().value
            )
            environment: str = (
                models.Configuration.query.filter_by(name="environment").one().value
            )

            if endpoint is not None:
                domain: Optional[models.GlobusDnsDomain] = endpoint.globus_dns_domain
                if domain:
                    node = domain.get_node(id=node_id)
                    if node:
                        try:
                            data_interface = node.data_interface
                        except AttributeError:
                            ip_addresses = node.ip_addresses
                            if ip_addresses:
                                data_interface = ip_addresses[0]

                        GridFTPService(
                            getattr(node, "incoming_port_range", None),
                            getattr(node, "outgoing_port_range", None),
                            data_interface,
                            domain.domain_name,
                            environment=environment,
                            port=endpoint.gridftp_control_channel_port,
                        ).update()

        vhost_dirty = False
        if changed_endpoint:
            # If endpoint changed, a subscription state change requires us
            # to replace all collection vhosts
            try:
                endpoint = models.Endpoint.query.one_or_none()

                if endpoint:
                    if endpoint.endpoint_id != self.app.config.get("HOST_ENDPOINT_ID"):
                        # Migrated uuid. Need to change some things:
                        #    endpoint_id in info.json
                        #    host_endpoint_id in the configuration table
                        #    current_app.config["HOST_ENDPOINT_ID"]
                        hei_config = models.Configuration.query.filter(
                            models.Configuration.name == "host_endpoint_id"
                        ).one()
                        hei_config.value = endpoint.endpoint_id
                        models.db.session.add(hei_config)
                        models.db.session.commit()
                        self.app.config["HOST_ENDPOINT_ID"] = endpoint.endpoint_id

                        # info.json will be updated in the call to endpoint.write_config()

                    self.app.subscription = getattr(endpoint, "subscription_id", None)
                    endpoint.write_config()
                    vhost_dirty = True
            except Exception as e:
                self.app.logger.warning("Unable to update the endpoint: %s", str(e))

        if changed_certificate:
            # If cert changed, need to update the local config
            try:
                certificate: Optional[
                    models.LetsEncryptCertificate
                ] = models.LetsEncryptCertificate.query.one_or_none()
                if certificate:
                    certificate.write_config(certificate.deleted)
                    vhost_dirty = True
            except Exception:
                self.app.logger.warning("Unable to update %s", certificate)

        collection_query = models.Collection.query
        if not changed_endpoint:
            # If managed state has not changed, only need to update vhosts for changed collections.
            # Note that hard-deleted collections will not show up in this query; fortunately, we
            # soft delete collections first to handle dependencies asynchronously.
            collection_query = collection_query.filter(
                models.Collection.id.in_(changed_collection_ids)
                | models.Collection.mapped_collection_id.in_(changed_collection_ids)  # type: ignore
            )

        collection: models.Collection
        for collection in collection_query:
            try:
                self.app.logger.debug(
                    f"{'Deleting' if collection.is_deleted else 'Updating'} collection vhost for %s",
                    collection,
                )
                # Checking is_deleted will force removal of guest vhosts when we soft delete
                # mapped collections.
                collection.write_config(collection.is_deleted)
                vhost_dirty = True
            except Exception as e:
                self.app.logger.warning(
                    "Unable to update %s vhost config: %s", collection, str(e)
                )

        if vhost_dirty:
            with self.mutex:
                self.vhost_dirty = True
                self.vhost_timestamp = datetime.datetime.now()

    def on_message(self, data) -> None:
        """
        Appsync change notification message handler.

        This function is called by the websocket processor to indicate a change
        notification message has been received. It deserializes ``data`` and
        passes it to ``NotificationManager.process_items`` to process the data
        or schedule it for deferred processing.

        :param data: The notification message from appsync.

        NOTE: This callback will be called within the current_app context.
        """
        with self.__subscription_lock.exclusive():
            old_subscription: Optional[str] = None
            try:
                endpoint: Optional[
                    models.Endpoint
                ] = models.Endpoint.query.one_or_none()
                if endpoint:
                    old_subscription = endpoint.subscription_id
            except Exception as e:
                self.app.logger.warning("Exception reading endpoint %s", str(e))

            result = {}
            with suppress(KeyError):
                result = data["data"]["Endpoint"]

            if result:
                parsed = appsync_data_schema.load(result, unknown=ma.EXCLUDE)
                self.app.logger.debug(
                    "Updates from appsync for %s",
                    [i["tag"].value + i["id"] for i in parsed["items"]],
                )

                self.process_items(
                    parsed["items"], old_subscription, ignore_old_updates=True
                )

                with self.mutex:
                    if self.vhost_dirty and not self.vhost_reload:
                        self.vhost_reload = self.scheduler.enter(
                            1, 0, self.touch_reload
                        )
            else:
                self.app.logger.warning("Couldn't parse data %s", data)

    def on_disconnect(self, *args, **kwargs) -> None:
        """
        Appsync subscription disconnect event handler.

        This function is called when appsync stops sending responses to the
        ping requests for this objects notifier object. When this occurs,
        we will synchronize and resubscribe.

        When disconnected, schedule the
        ``NotificationManager.query_and_subscribe`` method to run to update
        state and resume the subscription.

        NOTE: This callback will be called within the current_app context. It
        is called whenever the subscription event loop exits which will happen
        any time an error is encountered.
        """
        self.notifier = None
        self.scheduler.enter(self.ERROR_DELAY, 0, self.query_and_subscribe)

    def cleanup_domains_and_vhosts(self) -> None:
        """
        Remove any unused domains from the domains directory, so that
        the gridftp server doesn't get confused.
        """
        custom_domains = []
        vhosts = []

        collection: models.Collection
        for collection in models.Collection.query.filter(
            ~models.Collection.is_deleted  # type: ignore
        ):
            if collection.has_domain:
                custom_domains.append(collection.domain.domain_name)
            vhosts.append(collection.id + ".conf")

        ep: models.Endpoint = models.Endpoint.query.one()
        vhosts.append(os.path.basename(ep.vhost_config_path))

        if ep.has_domain:
            custom_domains.append(ep.domain.domain_name)

        if ep.has_oidc_server:
            custom_domains.append(ep.oidc_server.domain.domain_name)
            vhosts.append(ep.oidc_server.client_id + ".conf")

        letsencrypt: models.LetsEncryptCertificate = (
            models.LetsEncryptCertificate.query.one()
        )
        custom_domains.append(letsencrypt.domain_name)

        for entry in os.listdir(models.Domain.domain_dir):
            if entry not in custom_domains:
                entry_path = os.path.join(models.Domain.domain_dir, entry)
                shutil.rmtree(entry_path, ignore_errors=True, onerror=None)

        for entry in os.listdir(self.vhost_dir):
            if entry not in vhosts:
                entry_path = os.path.join(self.vhost_dir, entry)
                if not os.path.isdir(entry_path):
                    self.app.logger.debug(
                        "Removing %s from the vhosts config dir", entry
                    )
                    try:
                        os.remove(entry_path)
                    except Exception:
                        pass

    def touch_reload(self) -> None:
        """
        Write to the file at self.RELOAD_PATH, triggering apache to
        reload vhosts. This will only happen if no vhost changes have happened
        in the last second; otherwise, it will wait a second and try again.
        """
        with self.app.app_context():
            with self.mutex:
                if self.vhost_dirty:
                    if (
                        datetime.datetime.now() - self.vhost_timestamp
                    ) > datetime.timedelta(seconds=1):
                        self.cleanup_domains_and_vhosts()
                        with open(self.RELOAD_PATH, "w") as f:
                            f.write("reload")
                        self.vhost_dirty = False
                        self.vhost_reload = None
                    else:
                        self.vhost_reload = self.scheduler.enter(
                            1, 0, self.touch_reload
                        )
                else:
                    self.vhost_reload = None
