"""
globus-connect-server endpoint setup
"""
import typing as t
import uuid

import click

from gcs_cli.parsing.command_decorator import globus_command
from gcs_cli.parsing.types import CommaDelimitedList, StringOrNull

from globus.manager.utils.keychain import NodeKey
from .utils.key_and_credentials import key_and_credentials_options, KeyCredsAvailable


def _is_id_required(kca: KeyCredsAvailable):
    """
    client_id is required if a deployment_key is provided.
    """
    return kca["key"]


def _is_secret_required(kca: KeyCredsAvailable):
    """
    secret is required if deploymenmt_key or secret are provided.
    """
    return kca["client_id"] or kca["key"]


@globus_command(
    "setup",
    short_help="Setup a Globus Connect Server endpoint",
    help="Setup a Globus Connect Server endpoint",
    disable_options=["format", "explicit_host"],
)
@click.argument("display_name")
@click.option(
    "--public/--private",
    "public",
    default=True,
    help="Set the endpoint to be public or private [default: public]",
    show_default=False,
)
@click.option(
    "--description",
    type=StringOrNull(),
    help="Description of the endpoint",
)
@click.option(
    "--info-link",
    type=StringOrNull(),
    help="Link to a web page with more information about the endpoint",
)
@click.option(
    "--contact-info",
    type=StringOrNull(),
    help="Other non-email contact information for the endpoint",
)
@click.option(
    "--contact-email",
    type=StringOrNull(),
    help="Email address of the support contact for this endpoint",
    required=True,
)
@click.option(
    "--organization",
    type=StringOrNull(),
    help="Organization that owns this endpoint",
    required=True,
)
@click.option(
    "--department",
    type=StringOrNull(),
    help="Department within the organization that runs the endpoint",
)
@click.option(
    "--keywords",
    type=CommaDelimitedList(),
    help="Comma separated list of keywords to help searches for the endpoint",
)
@key_and_credentials_options(
    id_required=_is_id_required,
    secret_required=_is_secret_required,
    key_exists=False,
    key_writable=True,
)
@click.option(
    "--project-id",
    metavar="PROJECT_ID",
    type=click.UUID,
    help=(
        "Auth project ID where the new authentication policy will be stored. "
        "Only required if you are an admin on 2 or more Auth projects."
    ),
)
@click.option(
    "--owner",
    "-o",
    help="Globus username of the owner of this endpoint. This user will be "
    "granted the owner role for this endpoint.",
    required=True,
)
@click.option(
    "--agree-to-letsencrypt-tos",
    help="Agree to the Let's Encrypt Terms of Service: "
    "https://letsencrypt.org/repository/",
    is_flag=True,
)
@click.option(
    "--dont-set-advertised-owner",
    help="Skip the interactive step to set the endpoint advertised owner",
    is_flag=True,
)
@click.option(
    "--development-appsync",
    help="Use the specified URL and client_id for the appsync service",
    metavar="APPSYNC_URL APPSYNC_CLIENT_ID",
    nargs=2,
    hidden=True,
)
def endpoint_setup(
    display_name: str,
    client_id: t.Optional[uuid.UUID],
    secret: t.Optional[str],
    deployment_key_path: str,
    deployment_key: t.Optional[NodeKey],
    project_id: t.Optional[uuid.UUID],
    owner: str,
    agree_to_letsencrypt_tos: bool,
    dont_set_advertised_owner: bool,
    contact_email: str,
    development_appsync: t.Optional[t.Tuple[str, str]] = None,
    **kwargs
) -> None:
    from globus.manager.scripts.endpoint_setup import endpoint_setup as es

    es(
        display_name,
        str(client_id) if client_id else None,
        secret,
        deployment_key_path,
        deployment_key,
        str(project_id) if project_id else None,
        owner,
        agree_to_letsencrypt_tos,
        dont_set_advertised_owner,
        contact_email,
        development_appsync,
        **kwargs
    )
