import re
from typing import Any, Dict, List, Optional

from flask import current_app

from globus.manager.constants import CollectionType, ConfigurationObjectTag
from globus.manager.api.schemas.api import collection_private_schema
import globus.manager.api.models as models


class _CollectionV53:
    """
    Base class of a collection exported from a GCSv5.3 installation. This
    maps the properties of the storage gateway into something which can be
    dumped using the v5.4 marshmallow schema. It also contains some private
    fields which do not directly map to v5.4.
    """

    tag = ConfigurationObjectTag.COLLECTION

    def __init__(
        self,
        connector_id: str,
        transfer_doc: Dict[str, Any],
        allow_guest_collections: bool,
        owner: str,
        user_credential: Optional[models.UserCredential],
        storage_gateway: models.StorageGateway,
        storage_gateway_root: str,
        identity_mapping: Optional[List[Dict[str, Any]]],
        **kwargs
    ):
        self.id = kwargs["id"]
        self.domain_name = kwargs.get("domain")
        self.connector_id = connector_id
        self.display_name = transfer_doc["display_name"]
        if user_credential:
            self.identity_id = user_credential.identity_id  # type: ignore
        else:
            self.identity_id = owner  # type: ignore
        self._identity_mapping = identity_mapping
        self.storage_gateway_id = storage_gateway.id
        self.storage_gateway = storage_gateway
        self.mapped_collection_id = None
        self.default_directory = transfer_doc["default_directory"]
        self.public = transfer_doc["public"]
        self.force_encryption = transfer_doc["force_encryption"]
        self.disable_verify = transfer_doc["disable_verify"]

        if "organization" in transfer_doc:
            self.organization = transfer_doc.get("organization")
        keywords_str = transfer_doc.get("keywords")
        if keywords_str:
            self.keywords = [s.strip() for s in keywords_str.split(",")]
        if transfer_doc.get("department"):
            self.description = transfer_doc.get("description")
        if transfer_doc.get("contact_email"):
            self.contact_email = transfer_doc.get("contact_email")
        if transfer_doc.get("contact_info"):
            self.contact_info = transfer_doc.get("contact_info")
        if transfer_doc.get("info_link"):
            self.info_link = transfer_doc.get("info_link")
        if user_credential:
            self.user_credential_id = user_credential.id
        self.user_credential = user_credential
        self.high_assurance = storage_gateway.high_assurance
        self.authentication_timeout_mins = storage_gateway.authentication_timeout_mins
        self.allow_guest_collections = allow_guest_collections
        self.deleted = False

        # These are stored, but are only used internally and are not visible
        # to the outside
        self._created_at = kwargs["created_at"]
        self._updated_at = kwargs["updated_at"]

        public_policies = kwargs.get("collection_policies", {}).get("public", {})
        private_policies = kwargs.get("collection_policies", {}).get("private", {})

        self.init_public_policies(**public_policies)
        self.init_private_policies(**private_policies)

        if self.collection_type == CollectionType.mapped:
            self.collection_base_path = re.sub(
                "//*", "/", "/".join([storage_gateway_root, kwargs["path"]])
            )
        else:
            self.collection_base_path = kwargs["path"]

    def init_public_policies(self, **policies):
        # This does not exist in v5.4, but can be used when synthesizing
        # mapped collections for storage gateways that have guest collections
        # but no mapped collections
        self.collection_type = CollectionType(policies["type"])
        if self.collection_type == CollectionType.guest:
            delattr(self, "allow_guest_collections")

    def init_private_policies(self, **policies):
        default_directory = policies.get("default_directory")
        if default_directory:
            self.default_directory = default_directory
        else:
            self.default_directory = "/"

    @property
    def manager_url(self):
        domain = models.GlobusDnsDomain.query.one().domain_name
        return "https://{}".format(domain)

    def update(self, appsync_data):
        self.appsync_data = appsync_data

    @property
    def configuration(self):
        configuration_hash = collection_private_schema.dump(self)
        return current_app.keychain.encrypt(configuration_hash)


class PosixCollectionV53(_CollectionV53):
    pass


class GoogleCloudStorageCollectionV53(_CollectionV53):
    def init_private_policies(self, google_cloud_storage_project: str = None, **kwargs):
        self.project = google_cloud_storage_project

        super().init_private_policies(**kwargs)


class GoogleDriveCollectionV53(_CollectionV53):
    pass


class S3CollectionV53(_CollectionV53):
    pass


class CephCollectionV53(_CollectionV53):
    pass


class BoxCollectionV53(_CollectionV53):
    pass


class BlackPearlCollectionV53(_CollectionV53):
    pass


def CollectionV53(
    connector_id: str,
    transfer_doc: Dict[str, Any],
    allow_guest_collections: bool,
    owner: str,
    user_credential: Optional[models.UserCredential],
    storage_gateway: models.StorageGateway,
    storage_gateway_root: str,
    identity_mapping: Optional[List[Dict[str, Any]]],
    **kwargs
) -> _CollectionV53:
    """
    Pseudo-constructor for the *CollectionV53 classes, which uses the
    connector id from the storage gateway definition document to select which
    type of object to create.
    """

    connector_map = {
        "7e3f3f5e-350c-4717-891a-2f451c24b0d4": BlackPearlCollectionV53,
        "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63": BoxCollectionV53,
        "1b6374b0-f6a4-4cf7-a26f-f262d9c6ca72": CephCollectionV53,
        "56366b96-ac98-11e9-abac-9cb6d0d9fd63": GoogleCloudStorageCollectionV53,
        "976cf0cf-78c3-4aab-82d2-7c16adbcc281": GoogleDriveCollectionV53,
        "145812c8-decc-41f1-83cf-bb2a85a2a70b": PosixCollectionV53,
        "7643e831-5f6c-4b47-a07f-8ee90f401d23": S3CollectionV53,
    }

    assert connector_id in connector_map

    return connector_map[connector_id](
        connector_id,
        transfer_doc,
        allow_guest_collections,
        owner,
        user_credential,
        storage_gateway,
        storage_gateway_root,
        identity_mapping,
        **kwargs
    )
