import json
from os.path import exists
from functools import wraps

from typing import Any, Callable, Literal, Optional, overload, TypedDict, Union

import click

from globus.manager.utils.keychain import NodeKey


_CTX_HIDDEN_KEY = "_gcs"


def _ctx_set_value(ctx: click.Context, key: str, value: Union[str, NodeKey]) -> None:
    setattr(ctx, f"{_CTX_HIDDEN_KEY}_{key}", value)


@overload
def _ctx_get_value(
    ctx: click.Context, key: Literal["client_id", "secret"]
) -> Optional[str]:
    ...


@overload
def _ctx_get_value(
    ctx: click.Context, key: Literal["deployment_key"]
) -> Optional[NodeKey]:
    ...


def _ctx_get_value(ctx: click.Context, key: str) -> Optional[Union[str, NodeKey]]:
    return getattr(ctx, f"{_CTX_HIDDEN_KEY}_{key}", None)


class _ProcessDeploymentKey(click.Option):
    """
    Process --deployment-key (or its default value), load the file if it exists and
    store away the (key, client_id, secret) if they exist within the file.
    """

    def handle_parse_result(
        self, ctx: click.Context, opts: dict[str, Any], args: list[str]
    ) -> tuple[Any, list[str]]:
        # Allow the super to verify the deployment key path
        retval = super().handle_parse_result(ctx, opts, args)
        deployment_key_path = retval[0]

        # If the key exists, load it and store extra values as needed
        if exists(deployment_key_path):
            deployment_key_contents: Optional[dict] = None
            with open(deployment_key_path, "r") as f:
                try:
                    deployment_key_contents = json.load(f)
                except Exception as e:
                    raise click.ClickException(
                        f"Invalid key file '{deployment_key_path}': {e}"
                    )

            if deployment_key_contents:
                # Store away client_id and secret, if available in the key
                for key in ["client_id", "secret"]:
                    if key in deployment_key_contents:
                        _ctx_set_value(ctx, key, deployment_key_contents.pop(key))

                # Turn the remaining contents into a NodeKey and store it away
                try:
                    deployment_key = NodeKey(**deployment_key_contents)
                    _ctx_set_value(ctx, "deployment_key", deployment_key)
                except Exception as e:
                    raise click.ClickException(
                        f"Invalid key file '{deployment_key_path}': {e}"
                    )

        return retval


class KeyCredsAvailable(TypedDict):
    client_id: bool
    secret: bool
    key: bool


class _ProcessOption(click.Option):
    def __init__(
        self,
        *args,
        check_required: Union[bool, Callable[[KeyCredsAvailable], bool]] = False,
        **kwargs,
    ) -> None:
        self._check_required = check_required
        super().__init__(*args, **kwargs)

    def is_required(
        self,
        ctx: click.Context,
        opts: dict[str, Any],
    ) -> bool:
        if isinstance(self._check_required, bool):
            return self._check_required

        return self._check_required(
            KeyCredsAvailable(
                client_id=bool(_ctx_get_value(ctx, "client_id")) or "client_id" in opts,
                secret=bool(_ctx_get_value(ctx, "secret")) or "secret" in opts,
                key=bool(_ctx_get_value(ctx, "deployment_key")),
            )
        )


class _ProcessClientID(_ProcessOption):
    """
    If _ProcessDeploymentKey saved a client_id value from the deployment key, verify
    that it matches any client_id value on the command line or, if not given on the
    command line, inject client_id into opts so that click does not error on missing
    a required argument and click will verify its value.
    """

    def handle_parse_result(
        self,
        ctx: click.Context,
        opts: dict[str, Any],
        args: list[str],
    ) -> tuple[Any, list[str]]:
        # Determine if this option should be required given all we know
        self.required = self.is_required(ctx, opts)

        stored_client_id: Optional[str] = _ctx_get_value(ctx, "client_id")
        if stored_client_id:
            if "client_id" in opts:
                # Verify that the values are the same
                if stored_client_id != opts["client_id"]:
                    raise click.ClickException(
                        "The value for --client-id does not match the value provided "
                        "in the deployment key. Please rerun this command without "
                        "--client-id."
                    )
            else:
                # Let's use the value stored in the deployment key
                opts["client_id"] = stored_client_id

        # Let the super class validate the supplied value
        return super().handle_parse_result(ctx, opts, args)


class _ProcessSecret(_ProcessOption):
    """
    If _ProcessDeploymentKey saved a secret value from the deployment key, verify
    that it matches any secret value on the command line or, if not given on the
    command line, inject secret into opts so that click does not prompt for it.
    """

    def prompt_for_value(self, ctx):
        """
        Skip prompting if we have disabled it.
        """
        if self._disable_prompt:
            return None
        return super().prompt_for_value(ctx)

    def handle_parse_result(
        self, ctx: click.Context, opts: dict[str, Any], args: list[str]
    ) -> tuple[Any, list[str]]:

        # Don't prompt when missing if the option is not required
        self._disable_prompt = not self.is_required(ctx, opts)

        stored_secret: Optional[str] = _ctx_get_value(ctx, "secret")
        if stored_secret:
            if "secret" in opts:
                # Verify that the values are the same
                if stored_secret != opts["secret"]:
                    raise click.ClickException(
                        "The value for --secret does not match the value provided "
                        "in the deployment key. Please rerun this command without "
                        "--secret."
                    )
            else:
                # Let's use the value stored in the deployment key
                opts["secret"] = stored_secret

        # Let the super class validate the supplied value
        return super().handle_parse_result(ctx, opts, args)


def key_and_credentials_options(
    func: Optional[Callable] = None,
    id_required: Union[bool, Callable[[KeyCredsAvailable], bool]] = True,
    secret_required: Union[bool, Callable[[KeyCredsAvailable], bool]] = True,
    key_exists: bool = True,
    key_writable: bool = True,
) -> Callable:
    """
    Adds the following options:
    - --client-id: click.types.UUIDParameterType
    - --secret: click.STRING
    - --deployment-key: click.Path

    Adds the following args to the function:
    - client_id: Optional[uuid.UUID]
    - secret: Optional[str]
    - deployment_key_path: str
    - deployment_key: Optional[NodeKey]

    id_required: Set the value for click.Option.required. If id_required is a callable,
    the called function can determine if the option is required.

    secret: Set the value for click.Option.required. If secret is a callable, the called
    function can determine if the option is required.

    Note: --deployment-key is an eager option so that it can be processed before
    client_id and secret, regardless of order of arguments on the command line. In
    order for it to not conflict with other eager arguments (ex. --help, --version),
    make sure --deployment-key comes after all other eager arguments.
    """

    def _wrapped(_func: Callable) -> Callable:
        @click.option(
            "--deployment-key",
            "-d",
            "deployment_key_path",
            help="Path for deployment key configuration.",
            type=click.Path(
                dir_okay=False,
                exists=key_exists,
                writable=key_writable,
                readable=True,
            ),
            default="deployment-key.json",
            is_eager=True,  # This option must be evaluated before --client-id or --secret
            cls=_ProcessDeploymentKey,
        )
        @click.option(
            "--client-id",
            "-c",
            help=(
                "Auth client ID for this Globus Connect Server endpoint. Required "
                "for endpoints created before version 5.4.61."
            ),
            type=click.types.UUIDParameterType(),
            required=id_required if isinstance(id_required, bool) else False,
            cls=_ProcessClientID,
            check_required=id_required,
        )
        @click.option(  # type: ignore # hidden/required
            "--secret",
            "-s",
            help=(
                "Auth client secret for this Globus Connect Server endpoint. Required "
                "for endpoints created before version 5.4.61."
            ),
            hidden=True,
            prompt=True,
            hide_input=True,
            type=click.types.STRING,
            required=secret_required if isinstance(secret_required, bool) else False,
            cls=_ProcessSecret,
            check_required=secret_required,
        )
        @click.pass_context
        @wraps(_func)
        def wrapper(*args, **kwargs):
            assert len(args) == 1
            assert type(args[0]) == click.Context
            ctx = args[0]
            deployment_key = _ctx_get_value(ctx, "deployment_key")
            return _func(**kwargs, deployment_key=deployment_key)

        return wrapper

    return _wrapped
